import copy

import pytest
import seutil as su


def test_set():
    trie = su.ds.trie.Trie()
    trie.set("key1")
    assert "key1" in trie
    assert trie["key1"] is True

    trie["another"] = 123
    assert trie["another"] == 123

    trie.set("efg", 456)
    assert trie["efg"] == 456


def test_set_exist_ok():
    trie = su.ds.trie.Trie()
    trie.set("key1")
    trie.set("key1", exist_ok=True)
    assert "key1" in trie

    with pytest.raises(KeyError):
        trie.set("key1", exist_ok=False)


def test_get():
    trie = su.ds.trie.Trie()
    trie["key1"] = 123
    assert trie.get("key1") == 123
    assert trie["key1"] == 123


def test_get_missing():
    trie = su.ds.trie.Trie()
    assert trie.get("key1") is None

    with pytest.raises(KeyError):
        trie["key1"]

    with pytest.raises(KeyError):
        trie[""]


def test_get_default():
    trie = su.ds.trie.Trie()
    trie["key1"] = 123
    assert trie.get("key1", default=456) == 123
    assert trie.get("key2", default=456) == 456


def test_compute_add():
    trie = su.ds.trie.Trie()
    trie.compute("key1", lambda v: 123)
    assert trie["key1"] == 123


def test_compute_remove():
    trie = su.ds.trie.Trie()
    trie["key1"] = 123
    trie.compute("key1", lambda v: su.ds.trie.Trie.MISSING)
    assert "key1" not in trie


def test_compute_update():
    trie = su.ds.trie.Trie()
    trie["key1"] = 123
    trie.compute("key1", lambda v: 456)
    assert trie["key1"] == 456


def test_remove():
    trie = su.ds.trie.Trie()
    trie["key1"] = 123
    del trie["key1"]
    assert "key1" not in trie


def test_remove_missing():
    trie = su.ds.trie.Trie()
    with pytest.raises(KeyError):
        del trie["key1"]

    with pytest.raises(KeyError):
        del trie[""]


def test_has_key():
    trie = su.ds.trie.Trie()
    trie["key1"] = 123
    assert "key1" in trie
    assert "key2" not in trie


def test_has_prefix():
    trie = su.ds.trie.Trie()
    trie["key1"] = 123
    assert trie.has_prefix("ke")
    assert not trie.has_prefix("what")


def test_copy():
    trie = su.ds.trie.Trie()
    trie["key1"] = 123
    trie["mutable1"] = [4, 5, 6]
    trie_copy = copy.copy(trie)
    assert trie_copy["key1"] == 123
    assert trie_copy["mutable1"] == [4, 5, 6]

    trie["key2"] = 456
    assert "key2" not in trie_copy

    trie["mutable1"].append(7)
    assert trie_copy["mutable1"] == [4, 5, 6, 7]


def test_deepcopy():
    trie = su.ds.trie.Trie()
    trie["key1"] = 123
    trie["mutable1"] = [4, 5, 6]
    trie_copy = copy.deepcopy(trie)
    assert trie_copy["key1"] == 123
    assert trie_copy["mutable1"] == [4, 5, 6]

    trie["key2"] = 456
    assert "key2" not in trie_copy

    trie["mutable1"].append(7)
    assert trie_copy["mutable1"] == [4, 5, 6]


def test_get_subtrie():
    trie = su.ds.trie.Trie()
    trie["key1"] = 123
    trie["key2"] = 456

    subtrie = trie.get_subtrie("key")
    assert subtrie["1"] == 123
    assert subtrie["2"] == 456


def test_get_subtrie_mismatch():
    trie = su.ds.trie.Trie()
    trie["key1"] = 123
    trie["key2"] = 456

    subtrie = trie.get_subtrie("whatever")
    assert len(subtrie) == 0


def test_get_subtrie_prefix_of():
    trie = su.ds.trie.Trie()
    trie["abc"] = 123
    trie["abcd"] = 456
    trie["abcdef"] = 789
    trie["def"] = 999

    subtrie = trie.get_subtrie_prefix_of("abcdef")
    assert subtrie["abc"] == 123
    assert subtrie["abcd"] == 456
    assert subtrie["abcdef"] == 789
    assert "def" not in subtrie


def test_get_subtrie_using_elems():
    trie = su.ds.trie.Trie()
    trie["abc"] = 1
    trie["abcd"] = 2
    trie["abcabc"] = 3
    trie["def"] = 4

    subtrie = trie.get_subtrie_using_elems("abc")
    assert subtrie["abc"] == 1
    assert subtrie["abcabc"] == 3
    assert "abcd" not in subtrie
    assert "def" not in subtrie


def test_subtries():
    trie = su.ds.trie.Trie()
    trie["a1"] = 1
    trie["a2"] = 2
    trie["b3"] = 3
    trie["c4"] = 4

    e2subtrie = {k: v for k, v in trie.subtries()}
    assert len(e2subtrie) == 3
    assert len(e2subtrie["a"]) == 2
    assert len(e2subtrie["b"]) == 1
    assert len(e2subtrie["c"]) == 1


def test_items():
    trie = su.ds.trie.Trie()
    trie["a1"] = 1
    trie["a2"] = 2
    trie["b3"] = 3
    trie["c4"] = 4

    e2v = {k: v for k, v in trie.items()}
    assert len(e2v) == 4
    assert e2v["a1"] == 1
    assert e2v["a2"] == 2
    assert e2v["b3"] == 3
    assert e2v["c4"] == 4


def test_keys():
    trie = su.ds.trie.Trie()
    trie["a1"] = 1
    trie["a2"] = 2
    trie["b3"] = 3
    trie["c4"] = 4

    keys = list(trie.keys())
    assert len(keys) == 4
    assert "a1" in keys
    assert "a2" in keys
    assert "b3" in keys
    assert "c4" in keys


def test_values():
    trie = su.ds.trie.Trie()
    trie["a1"] = 1
    trie["a2"] = 2
    trie["b3"] = 3
    trie["c4"] = 4

    values = list(trie.values())
    assert len(values) == 4
    assert 1 in values
    assert 2 in values
    assert 3 in values
    assert 4 in values


def test_overlaps_with():
    trie1 = su.ds.trie.Trie()
    trie1["abc"] = 1
    trie1["def"] = 2

    trie2 = su.ds.trie.Trie()
    trie2["ab"] = 3
    trie2["defg"] = 4

    e2overlaps = {k: (t1, t2) for k, t1, t2 in trie1.overlaps_with(trie2)}
    assert set(e2overlaps.keys()) == {"", "a", "ab", "d", "de", "def"}


def test_overlaps_with_self_has_value():
    trie1 = su.ds.trie.Trie()
    trie1["alpha"] = 1
    trie1["beta"] = 2
    trie1["gamma"] = 3

    trie2 = su.ds.trie.Trie()
    trie2["alphaXXX"] = 1
    trie2["betaXXX"] = 1

    e2overlaps = {
        k: (t1, t2) for k, t1, t2 in trie1.overlaps_with(trie2, self_has_value=True)
    }
    assert len(e2overlaps) == 2
    assert e2overlaps["alpha"][0][""] == 1
    assert "" not in e2overlaps["alpha"][1]
    assert e2overlaps["beta"][0][""] == 2
    assert "" not in e2overlaps["beta"][1]


def test_overlaps_with_other_has_value():
    trie1 = su.ds.trie.Trie()
    trie1["alpha"] = 1
    trie1["beta"] = 2
    trie1["gamma"] = 3

    trie2 = su.ds.trie.Trie()
    trie2["aaa"] = 4
    trie2["b"] = 5

    e2overlaps = {
        k: (t1, t2) for k, t1, t2 in trie1.overlaps_with(trie2, other_has_value=True)
    }
    assert len(e2overlaps) == 1
    assert e2overlaps["b"][1][""] == 5
    assert "" not in e2overlaps["b"][0]



"""
Adding simple implementations or interfaces to some data types that are missing from the standard library.
"""

import lazy_import

from .graph_common import (Edge, EdgeExistedError, InvariantError, Node,
                           NodeIndexError)

lattice = lazy_import.lazy_module("seutil.ds.lattice")
trie = lazy_import.lazy_module("seutil.ds.trie")



from .graph_common import GraphLikeADT, Node
from typing import List, Optional, Set


class Lattice(GraphLikeADT):
    """
    Lattice: directed, acyclic, potentially connected graph.
    """

    def __init__(self, connected: bool = False):
        super().__init__(directed=True)
        self.connected = connected
        self.entry_nodes: Set[Node] = set()
        self.exit_nodes: Set[Node] = set()

    def check_invariants(self) -> bool:
        if self.g.vcount() == 0:
            return True
        return self.g.is_dag() and ((not self.connected) or self.g.is_connected("weak"))

    def add_node(
        self,
        parents: Optional[List[Node]] = None,
        children: Optional[List[Node]] = None,
    ) -> Node:
        node = super().add_node(parents, children)
        if parents is None:
            self.entry_nodes.add(node)
        else:
            for n in parents:
                if n in self.exit_nodes:
                    self.exit_nodes.remove(n)
        if children is None:
            self.exit_nodes.add(node)
        else:
            for n in children:
                if n in self.entry_nodes:
                    self.entry_nodes.remove(n)
        return node



import os
import subprocess
import warnings
from typing import Optional

from seutil import io


class BashError(RuntimeError):
    def __init__(
        self,
        cmd: str,
        completed_process: subprocess.CompletedProcess,
        check_returncode: int,
    ):
        self.cmd = cmd
        self.returncode = completed_process.returncode
        self.check_returncode = check_returncode
        self.stdout = completed_process.stdout
        self.stderr = completed_process.stderr

    def __str__(self) -> str:
        s = f"Command '{self.cmd}' failed with returncode {self.returncode}, expected returncode {self.check_returncode}.\n"
        show_full_output = os.environ.get("SEUTIL_SHOW_FULL_OUTPUT", "0") == "1"
        if show_full_output:
            s += f"STDOUT:\n{self.stdout}\n"
            s += f"STDERR:\n{self.stderr}\n"
        else:
            if len(self.stdout) > 200:
                s += f"STDOUT (truncated, set SEUTIL_SHOW_FULL_OUTPUT=1 to see full output):\n{self.stdout[:100]}...{self.stdout[-100:]}\n"
            else:
                s += f"STDOUT:\n{self.stdout}\n"
            if len(self.stderr) > 200:
                s += f"STDERR (truncated, set SEUTIL_SHOW_FULL_OUTPUT=1 to see full output):\n{self.stderr[:100]}...{self.stderr[-100:]}\n"
            else:
                s += f"STDERR:\n{self.stderr}\n"
        return s

    def __repr__(self) -> str:
        return self.__str__()


def run(
    cmd: str,
    check_returncode: Optional[int] = None,
    warn_nonzero: bool = True,
    update_env: bool = False,
    update_env_clear_existing: bool = False,
    **kwargs,
) -> subprocess.CompletedProcess:
    """
    Run a bash command using subprocess.run.  The command will be run using "bash -c".

    Some arguments' default values are changed (but can be overridden with kwargs):
    * capture_output=True, text=True:  capture all stdout and stderr.

    This function is able to check if return code match a given value (subprocess only supports
    checking non-zero values, but this function supports any).  Nevertheless, this function
    warns about any non-zero values if check_returncode is not set, to avoid silent failures;
    this behavior can be turned off via warn_nonzero=False.

    In addition, this function can try to update the environment variables in this process
    with the ones after running the command (if the command finished successfully).
    The retrival of the sub shell's environments is done by `env` into a temporary file.

    :param cmd: the command to run
    :param check_returncode: the return code to expect from the command
    :param warn_nonzero: whether to warn about non-zero exit codes
    :param update_env: whether to update the environment variables in this process
    :param update_env_clear_existing: whether to clear existing environment variables before updating
    :param kwargs: other arguments passed to subprocess.run
    :return: the subprocess.CompletedProcess object, has stdout, stderr, returncode fields
    :raises: BashError if the command's output did not match check_returncode
    :raises: subprocess.TimeoutExpired if the command timed out
    """
    # If update env is requested, append `env` to command
    if update_env:
        tempfile_update_env = io.mktmp("seutil-bash", ".txt")
        cmd += f" ; env > {tempfile_update_env}"

    kwargs.setdefault("capture_output", True)
    kwargs.setdefault("text", True)

    completed_process = subprocess.run(
        ["bash", "-c", cmd],
        **kwargs,
    )

    # check return code
    if (
        check_returncode is not None
        and completed_process.returncode != check_returncode
    ):
        raise BashError(cmd, completed_process, check_returncode)

    if (
        completed_process.returncode != 0
        and check_returncode is not None
        and warn_nonzero
    ):
        warnings.warn(
            f"Bash command `{cmd}` returned non-zero exit code: {completed_process.returncode}",
            RuntimeWarning,
        )

    if update_env:
        envs = io.load(tempfile_update_env, io.Fmt.txtList)
        if update_env_clear_existing:
            os.environ.clear()
        for env in envs:
            key, value = env.split("=", 1)
            os.environ[key] = value
        io.rm(tempfile_update_env)

    return completed_process



import os
import sys

module_root = os.path.dirname(os.path.realpath(__file__)) + "/.."
if module_root not in sys.path:
    sys.path.insert(0, module_root)

import lazy_import
from . import arg, bash, debug, io, log, pbar, project
from .BashUtils import BashUtils
from .GitHubUtils import GitHubUtils
from .IOUtils import IOUtils
from .LoggingUtils import LoggingUtils
from .Stream import Stream
from .TimeUtils import TimeoutException, TimeUtils
ds = lazy_import.lazy_module("seutil.ds")

__all__ = [
    # Classes
    "BashUtils",
    "CliUtils",
    "GitHubUtils",
    "IOUtils",
    "LoggingUtils",
    "MiscUtils",
    "Stream",
    "TimeUtils",
    # Sub-Packages
    "latex",
    # Exceptions
    "TimeoutException",
]

# Remove temporary names
del os
del sys
del module_root



import os
import sys

module_root = os.path.dirname(os.path.realpath(__file__)) + "/../.."
if module_root not in sys.path:
    sys.path.insert(0, module_root)

from .File import File
from .Macro import Macro

__all__ = [
    "File",
    "Macro",
]



import collections
import dataclasses
import inspect
import io
import json
import os
import pickle as pkl
import pydoc
import shutil
import tempfile
from enum import Enum
from pathlib import Path
from typing import (
    Any,
    Callable,
    Iterator,
    List,
    NamedTuple,
    Optional,
    Type,
    Union,
    get_type_hints,
)

import recordclass
import typing_inspect
import yaml

__all__ = [
    "cd",
    "rmdir",
    "rm",
    "mkdir",
    "mktmp",
    "mktmp_dir",
    "Fmt",
    "serialize",
    "deserialize",
    "load",
    "dump",
    "DeserializationError",
]


# ==========
# private utility functions
# ==========


def _unify_path(path: Union[str, Path]) -> Path:
    if not isinstance(path, Path):
        path = Path(path)
    return path


def _is_obj_record_class(obj: Any) -> bool:
    return (
        obj is not None
        and isinstance(obj, recordclass.mutabletuple)
        or isinstance(obj, recordclass.dataobject)
    )


def _is_clz_record_class(clz: Type) -> bool:
    return (
        clz is not None
        and inspect.isclass(clz)
        and (
            issubclass(clz, recordclass.mutabletuple)
            or issubclass(clz, recordclass.dataobject)
        )
    )


def _is_obj_named_tuple(obj: Any) -> bool:
    return obj is not None and isinstance(obj, tuple) and hasattr(obj, "_fields")


def _is_clz_named_tuple(clz: Type) -> bool:
    return (
        clz is not None
        and inspect.isclass(clz)
        and issubclass(clz, tuple)
        and hasattr(clz, "_fields")
    )


# ==========
# file and directory manipulation (creation/removal/browsing)
# ==========


class cd:
    """
    Temporally changes directory, for use with `with`:

    ```
    with cd(path):
        # cwd moved to path
        <statements>
    # cwd moved back to original cwd
    ```
    """

    def __init__(self, path: Union[str, Path]):
        path = _unify_path(path)
        self.path = path  # Path
        self.old_path = Path.cwd()  # Path

    def __enter__(self):
        os.chdir(self.path)

    def __exit__(self, type, value, tb):
        os.chdir(self.old_path)


def rmdir(
    path: Union[str, Path],
    missing_ok: bool = True,
    force: bool = True,
):
    """
    Removes a directory.

    :param path: the name of the directory.
    :param missing_ok: (-f) ignores error if the directory does not exist.
    :param force: (-f) force remove the directory even it's non-empty.
    """
    path = _unify_path(path)
    if path.is_dir():
        if force:
            shutil.rmtree(path, ignore_errors=True)
        else:
            path.rmdir()
    else:
        if path.exists():
            raise OSError(f"Use rm to remove regular file {path}")
        else:
            if missing_ok:
                return
            else:
                raise FileNotFoundError(f"Cannot remove non-exist directory {path}")


def rm(
    path: Union[str, Path],
    missing_ok: bool = True,
    force: bool = True,
):
    """
    Removes a file/directory.

    :param path: the name of the file/directory.
    :param missing_ok: (-f) ignores error if the file/directory does not exist.
    :param force: (-rf) force remove the directory even it's not empty.
    """
    path = _unify_path(path)
    if path.is_dir():
        rmdir(path, missing_ok=missing_ok, force=force)
    elif path.exists():
        path.unlink(missing_ok=missing_ok)
    else:
        if not missing_ok:
            raise FileNotFoundError(f"Cannot remove non-exist file {path}")


def mkdir(
    path: Union[str, Path],
    parents: bool = True,
    fresh: bool = False,
):
    """
    Creates a directory.

    :param path: the path to the directory.
    :param parents: if True, automatically creates parent directories; otherwise, raise error if any parent is missing.
    :param fresh: if True and if the directory already exists, removes it before creating.
    """
    path = _unify_path(path)

    if path.exists() and fresh:
        rmdir(path)

    path.mkdir(parents=parents, exist_ok=not fresh)


def mktmp(
    prefix: Optional[str] = None,
    suffix: Optional[str] = None,
    separator: str = "-",
    dir: Optional[Path] = None,
) -> Path:
    """
    Makes a temp file.  A wrapper for `tempfile.mkstemp`.
    """
    if prefix is not None:
        prefix = prefix + separator
    if suffix is not None:
        suffix = separator + suffix
    _, path = tempfile.mkstemp(prefix=prefix, suffix=suffix, dir=dir)
    return Path(path)


def mktmp_dir(
    prefix: Optional[str] = None,
    suffix: Optional[str] = None,
    separator: str = "-",
    dir: Optional[Path] = None,
) -> Path:
    """
    Makes a temp directory.  A wrapper for `tempfile.mkdtemp`.
    """
    if prefix is not None:
        prefix = prefix + separator
    if suffix is not None:
        suffix = separator + suffix
    path = tempfile.mkdtemp(prefix=prefix, suffix=suffix, dir=dir)
    return Path(path)


# ==========
# multi-format dumping/loading with serialization
# ==========


class FmtProperty(NamedTuple):
    # The function used by dump
    # * list_like=False:  takes a file-object and obj as input, writes the obj to the file-object
    # * list_like=True:  takes an item in the obj as input (from for loop), returns one line of text *without* "\n"
    writer: Union[Callable[[io.IOBase, Any], None], Callable[[Any], str]]

    # The function used by load
    # * list_like=False:  takes a file-object as input, reads the entire file and returns the obtained obj
    # * list_like=True:  takes one line of text as input, returns the obtained obj
    reader: Union[Callable[[io.IOBase], Any], Callable[[str], Any]]

    # File extensions, used for format inference; the first extension is used for output
    exts: List[str] = None

    # If the file should be opened in binary mode
    binary: bool = False

    # If the file should be read/writen one line at a time
    line_mode: bool = False

    # If this format requires (de)serialization
    serialize: bool = False


class Fmt(FmtProperty, Enum):
    # === txt ===
    txt = FmtProperty(
        writer=lambda f, obj: f.write(str(obj)),
        reader=lambda f: f.read(),
        exts=["txt"],
    )
    # === pickle ===
    pickle = FmtProperty(
        writer=lambda f, obj: pkl.dump(obj, f),
        reader=lambda f: pkl.load(f),
        exts=["pkl", "pickle"],
        binary=True,
    )
    # === json ===
    json = FmtProperty(
        writer=lambda f, obj: json.dump(obj, f, sort_keys=True),
        reader=lambda f: json.load(f),
        exts=["json"],
        serialize=True,
    )
    # Use yaml loader to allow formatting errors (e.g., trailing commas), but cannot handle unprintable chars
    jsonFlexible = json._replace(reader=lambda f: yaml.load(f, Loader=yaml.FullLoader))
    # Pretty-print version with sorting keys
    jsonPretty = json._replace(
        writer=lambda f, obj: json.dump(obj, f, sort_keys=True, indent=4),
    )
    # Pretty-print version without sorting keys
    jsonNoSort = json._replace(
        writer=lambda f, obj: json.dump(obj, f, indent=4),
    )
    # === jsonl (json list) ===
    jsonList = FmtProperty(
        writer=lambda item: json.dumps(item),
        reader=lambda line: json.loads(line),
        exts=["jsonl"],
        line_mode=True,
        serialize=True,
    )
    # === text list ===
    txtList = FmtProperty(
        writer=lambda item: str(item),
        reader=lambda line: line.replace("\n", ""),
        exts=["txt"],
        line_mode=True,
    )
    # === yaml ===
    yaml = FmtProperty(
        writer=lambda f, obj: yaml.dump(obj, f),
        reader=lambda f: yaml.load(f, Loader=yaml.FullLoader),
        exts=["yml", "yaml"],
        serialize=True,
    )


def infer_fmt_from_ext(ext: str, default: Optional[Fmt] = None) -> Fmt:
    if ext.startswith("."):
        ext = ext[1:]

    for fmt in Fmt:
        if fmt.exts is not None and ext in fmt.exts:
            return fmt

    if default is not None:
        return default
    else:
        raise RuntimeError(f'Cannot infer format for extension "{ext}"')


def serialize(
    obj: object,
    fmt: Optional[Fmt] = None,
) -> object:
    """
    Serializes an object into a data structure with only primitive types, list, dict.
    If fmt is provided, its formatting constraints are taken into account. Supported fmts:
    * json, jsonPretty, jsonNoSort, jsonList: dict only have str keys.

    :param obj: the object to be serialized.
    :param fmt: (optional) the target format.
    :return: the serialized object.
    """
    # Examine the type of object and use the appropriate serialization method
    # Check for simple types first
    if obj is None:
        return None
    elif isinstance(obj, (int, float, str, bool)):
        # Primitive types: keep as-is
        return obj
    elif hasattr(obj, "serialize"):
        # Call customized serialization method if exists
        return getattr(obj, "serialize")()
    elif _is_obj_named_tuple(obj):
        # NamedTuple
        return {k: serialize(v, fmt) for k, v in obj._asdict().items()}
    elif dataclasses.is_dataclass(obj):
        # Dataclass
        return {f.name: serialize(getattr(obj, f.name), fmt) for f in dataclasses.fields(obj)}
    elif isinstance(obj, (list, set, tuple)):
        # List-like: uniform to list; recursively serialize content
        return [serialize(item, fmt) for item in obj]
    elif isinstance(obj, dict):
        # Dict: recursively serialize content
        ret = {serialize(k, fmt): serialize(v, fmt) for k, v in obj.items()}

        # Json-like formats constraint: dict key must be str
        if fmt in [Fmt.json, Fmt.jsonPretty, Fmt.jsonNoSort, Fmt.jsonList]:
            ret = {str(k): v for k, v in ret.items()}
        return ret
    elif isinstance(obj, Enum):
        # Enum: use name
        return serialize(obj.name, fmt)
    elif _is_obj_record_class(obj):
        # RecordClass: convert to dict
        if hasattr(obj, "__dict__"):
            # Older versions of recordclass
            return {k: serialize(v, fmt) for k, v in obj.__dict__.items()}
        else:
            # Newer versions of recordclass
            return {f: serialize(getattr(obj, f), fmt) for f in obj.__fields__}
    else:
        raise TypeError(
            f"Cannot serialize object of type {type(obj)}, please consider writing a serialize() function"
        )

    # TODO: handle numpy arrays, pandas structures, pytorch structures, etc.


class DeserializationError(RuntimeError):
    def __init__(self, data, clz: Optional[Union[Type, str]], reason: str):
        self.data = data
        self.clz = clz
        self.reason = reason

    def __str__(self):
        return f"Cannot deserialize the following data to {self.clz}: {self.reason}\n  {self.data}"


_NON_TYPE = type(None)


def deserialize(
    data,
    clz: Optional[Union[Type, str]] = None,
    error: str = "ignore",
):
    """
    Deserializes some data (with only primitive types, list, dict) to an object with
    proper types.

    :param data: the data to be deserialized.
    :param clz: the targeted type of deserialization (or its name); if None, will return
        the data as-is.
    :param error: what to do when the deserialization has problem:
        * raise: raise a DeserializationError.
        * ignore (default): return the data as-is.
    :return: the deserialized data.
    """
    if clz is None:
        return data

    assert error in ["raise", "ignore"]

    # Resolve type by name
    # TODO: cannot resolve generic types
    if isinstance(clz, str):
        clz = pydoc.locate(clz)

    # NoneType
    if clz == _NON_TYPE:
        if data is None:
            return data
        else:
            raise DeserializationError(data, clz, "None type received non-None data")

    clz_origin = typing_inspect.get_origin(clz)
    if clz_origin is None:
        clz_origin = clz
        generic = False
    else:
        generic = True
    clz_args = typing_inspect.get_args(clz)

    # print(f"deserialize({data=}, {clz=}, {error=}), {clz_origin=}, {clz_args=}")

    # Optional type: extract inner type
    if typing_inspect.is_optional_type(clz):
        if data is None:
            return None
        inner_clz = clz_args[0]
        try:
            return deserialize(data, inner_clz, error=error)
        except DeserializationError as e:
            raise DeserializationError(data, clz, f"(Optional removed) " + e.reason)

    # Union type: try each inner type
    if typing_inspect.is_union_type(clz):
        ret = None
        for inner_clz in clz_args:
            try:
                ret = deserialize(data, inner_clz, error="raise")
            except DeserializationError:
                continue

        if ret is None:
            if error == "raise":
                raise DeserializationError(
                    data, clz, "All inner types are incompatible"
                )
            else:
                return data
        else:
            return ret

    # None data, but not NoneType
    if data is None:
        if error == "raise":
            raise DeserializationError(data, clz, "None data for non-None type")
        else:
            return data

    # List-like types
    if clz_origin in [list, tuple, set, collections.deque, frozenset]:
        if not isinstance(data, list):
            if error == "raise":
                raise DeserializationError(
                    data, clz, "Data does not have list structure"
                )
            else:
                return data

        if clz_origin == tuple:
            # Unpack list to tuple
            return tuple(
                [
                    # If the list has more items than Tuple[xxx] declared (e.g., [1, 2, 3], Tuple[int]), repeat the last declared type
                    deserialize(
                        x,
                        clz_args[min(i, len(clz_args) - 1)] if generic else None,
                        error=error,
                    )
                    for i, x in enumerate(data)
                ]
            )
        else:
            # Unpack list
            ret = [
                deserialize(x, clz_args[0] if generic else None, error=error)
                for x in data
            ]

            if clz_origin != list:
                # Convert to appropriate type
                return clz_origin(ret)
            else:
                return ret

    # Dict-like types
    if clz_origin in [
        dict,
        collections.OrderedDict,
        collections.defaultdict,
        collections.Counter,
    ]:
        if not isinstance(data, dict):
            if error == "raise":
                raise DeserializationError(
                    data, clz, "Data does not have dict structure"
                )
            else:
                return data

        if clz_origin == collections.OrderedDict:
            raise RuntimeWarning(
                f"The order of items in OrderedDict may not be preserved"
            )

        # Unpack dict
        ret = {
            deserialize(k, clz_args[0] if generic else None, error=error): deserialize(
                v, clz_args[1] if generic else None, error=error
            )
            for k, v in data.items()
        }
        if clz_origin != dict:
            # Convert to appropriate type
            return clz_origin(ret)
        else:
            return ret

    # Use customized deserialize function, if exists
    if inspect.isclass(clz) and hasattr(clz, "deserialize"):
        # TODO: check parameter of the deserialize function
        return getattr(clz, "deserialize")(data)

    # Enum
    if inspect.isclass(clz) and issubclass(clz, Enum):
        if isinstance(data, str):
            return clz[data]
        else:
            if error == "raise":
                raise DeserializationError(data, clz, "Enum data must be str (name)")
            else:
                return data

    # RecordClass
    if _is_clz_record_class(clz):
        field_values = {}
        for f, t in get_type_hints(clz).items():
            if f in data:
                field_values[f] = deserialize(data.get(f), t, error=error)
        return clz(**field_values)

    # NamedTuple
    if _is_clz_named_tuple(clz):
        field_values = {}
        for f in clz._fields:
            if hasattr(clz, "_field_types"):
                # for Python <3.9
                t = clz._field_types.get(f)
            elif hasattr(clz, '__annotations__'):
                # for Python >=3.9
                t = clz.__annotations__.get(f)
            else:
                t = None
            if f in data:
                field_values[f] = deserialize(data.get(f), t, error=error)
        return clz(**field_values)

    # DataClass
    if dataclasses.is_dataclass(clz):
        init_field_values = {}
        non_init_field_values = {}
        for f in dataclasses.fields(clz):
            if f.name in data:
                field_values = init_field_values if f.init else non_init_field_values
                field_values[f.name] = deserialize(
                    data.get(f.name), f.type, error=error
                )
        obj = clz(**init_field_values)
        for f_name, f_value in non_init_field_values.items():
            # use object.__setattr__ in case clz is frozen
            object.__setattr__(obj, f_name, f_value)
        return obj

    # Primitive types
    if clz_origin == type(data):
        return data
    if clz_origin == float and type(data) == int:
        return data

    if error == "raise":
        raise DeserializationError(
            data,
            clz,
            f"Cannot match requested type ({clz} / {clz_origin}) with data's type ({type(data)})",
        )
    else:
        return data


def dump(
    path: Union[str, Path],
    obj: object,
    fmt: Optional[Fmt] = None,
    serialization: Optional[bool] = None,
    parents: bool = True,
    append: bool = False,
    exists_ok: bool = True,
    serialization_fmt_aware: bool = True,
) -> None:
    """
    Saves an object to a file.
    The format is automatically inferred from the file name, if not otherwise specified.
    By default, serialization (i.e., converting to primitive types and data structures) is
    automatically performed for the formats that needs it (e.g., json).

    :param path: the path to save the file.
    :param obj: the object to be saved.
    :param fmt: the format of the file; if None (default), inferred from path.
    :param serialization: whether or not to serialize the object before saving:
        * True: always serialize;
        * None (default): only serialize for the formats that needs it;
        * False: never serialize.
    :param parents: what to do if parent directories of path do not exist:
        * True (default): automatically create them;
        * False: raise Exception.
    :param append: whether to append to an existing file if any (default False).
    :param exists_ok: what to do if path already exists and append is False:
        * True (default): automatically rewrites it;
        * False: raise Exception.
    :param serialization_fmt_aware: let the serialization function be aware of the target
        format to fit its constraints (e.g., dictionaries in json format can only have
        str keys).
    """
    path = _unify_path(path)

    # Check path existence
    if path.exists() and not exists_ok:
        raise FileExistsError(str(path))

    # Create parent directories
    if not path.parent.is_dir():
        if parents:
            path.parent.mkdir(parents=True)
        else:
            raise FileNotFoundError(str(path.parent))

    # Infer format
    if fmt is None:
        fmt = infer_fmt_from_ext(path.suffix)

    if append and not fmt.line_mode:
        raise RuntimeWarning(
            f"Appending to a format that's not list-like ({fmt}) may result in invalid file"
        )

    # Serialize (when appropriate)
    if serialization is None:
        serialization = fmt.serialize

    if serialization:
        obj = serialize(
            obj,
            fmt=fmt if serialization_fmt_aware else None,
        )

    # Open file
    file_mode = "w" if not append else "a"
    if fmt.binary:
        file_mode += "b"

    with open(path, file_mode) as f:
        # Write content
        if not fmt.line_mode:
            fmt.writer(f, obj)
        else:
            for item in obj:
                # Removing all "\n" inside the line
                f.write(fmt.writer(item).replace("\n", " ") + "\n")


def load(
    path: Union[str, Path],
    fmt: Optional[Fmt] = None,
    serialization: Optional[bool] = None,
    clz: Optional[Type] = None,
    error: str = "ignore",
    iter_line: bool = False,
) -> Union[object, Iterator[object]]:
    """
    Loads an object from a file.
    The format is automatically inferred from the file name, if not otherwise specified.
    By default, if clz is given, deserialization (i.e., unpackingn from primitive types
    and data structures) is automatically performed for the formats that needs it (e.g., json).

    :param path: the path to load the object.
    :param fmt: the format of the file; if None (default), inferred from path.
    :param serialization: whether or not to deserialize the object after loading:
        * True: always serialize;
        * None (default): only serialize for the formats that needs it;
        * False: never serialize.
    :param clz: the class to use for deserialization; if None (default), deserialization is a no-op.
    :param error: what to do if deserialization fails:
        * raise: raise a DeserializationError.
        * ignore (default): return the data as-is.
    :param iter_line: whether to iterate over the lines of the file instead of loading the whole file.
    """
    path = _unify_path(path)

    # Infer format
    if fmt is None:
        fmt = infer_fmt_from_ext(path.suffix)

    # Check arguments
    if iter_line and not fmt.line_mode:
        raise RuntimeError(f"Cannot load format {fmt} file under line mode")

    if serialization is None:
        serialization = fmt.serialize

    # Open file
    file_mode = "r"
    if fmt.binary:
        file_mode += "b"

    # Load content
    if not fmt.line_mode:
        with open(path, file_mode) as f:
            obj = fmt.reader(f)
            if serialization:
                obj = deserialize(obj, clz, error=error)
            return obj
    else:
        iterator = LoadIterator(path, file_mode, fmt, serialization, clz, error)
        if iter_line:
            return iterator
        else:
            return list(iterator)


class LoadIterator(Iterator):
    def __init__(
        self,
        path: Path,
        file_mode: str,
        fmt: FmtProperty,
        serialization: bool,
        clz: Optional[Type],
        error: str = "ignore",
    ):
        self.fd = open(path, file_mode)
        self.fmt = fmt
        self.serialization = serialization
        self.clz = clz
        self.error = error

    def __next__(self):
        line = self.fd.readline()
        if line == "":
            # EOF
            self.fd.close()
            raise StopIteration
        item = self.fmt.reader(line)
        if self.serialization:
            item = deserialize(item, self.clz, error=self.error)
        return item



class Table:

    def __init__(self, big=False, font_size="small", align="center", caption="Default Table",
                 label="fig:table:default"):
        self.big = big
        self.font_size = font_size
        self.align = align
        self.caption = caption
        self.label = label

        self.tabular = ""
        self.content = list()

    def __str__(self):
        latex_str = ""

        latex_str += "% Auto Generated From senl.util.latex.Table\n"

        if not self.big:
            latex_str += "\\begin{table}\n"
        else:
            latex_str += "\\begin{table*}\n"

        latex_str += "\\begin{{{0}}}\n".format(self.font_size)
        latex_str += "\\begin{{{0}}}\n".format(self.align)

        latex_str += "\\caption{{{0}}}\n".format(self.caption)
        latex_str += "\\label{{{0}}}\n".format(self.label)

        latex_str += "\\begin{{tabular}}{{{0}}}\n".format(self.tabular)

        for line in self.content:
            if isinstance(line, str):
                latex_str += line + "\n"
            else:
                latex_str += " & ".join(line) + " \\\\\n"

        latex_str += "\\end{tabular}\n"

        latex_str += "\\end{{{0}}}\n".format(self.align)
        latex_str += "\\end{{{0}}}\n".format(self.font_size)

        if not self.big:
            latex_str += "\\end{table}\n"
        else:
            latex_str += "\\end{table*}\n"

        return latex_str



import abc
from typing import Iterable, List, Optional

import igraph

"""
Base classes for graph-like data types (including trees and constrained graphs).

Currently we use igraph as the underlying graph library, but we add a level of abstraction to make it better fits our needs, and also make it possible to switch to other libraries in the future.
"""


class EdgeExistedError(ValueError):
    """Used to indicate that adding an edge failed because the edge already existed."""

    pass


class NodeIndexError(IndexError):
    """Used to indicate that a node index is out of range."""

    pass


class InvariantError(ValueError):
    """Used to indicate that a data structure invariant is violated."""

    pass


class Node:
    """
    Wraps an igraph.Vertex, but routes back all modification operations to the container class for checking the invariants.
    """

    def __init__(self, vertex: igraph.Vertex, container: "GraphLikeADT"):
        self.vertex = vertex
        self.container = container

    def __getitem__(self, key):
        if key == "name":
            raise IndexError("'name' attribute is reserved for internal use.")
        return self.vertex[key]

    def __setitem__(self, key, value):
        if key == "name":
            raise IndexError("'name' attribute is reserved for internal use.")
        self.vertex[key] = value

    def delete(self):
        self.container.delete_node(self.vertex["name"])

    @property
    def key(self):
        return self.vertex["name"]


class Edge:
    """
    Wraps an igraph.Edge, but routes back all modification operations to the container class for checking the invariants.
    """

    def __init__(self, edge: igraph.Edge, container: "GraphLikeADT"):
        self.edge = edge
        self.container = container


class GraphLikeADT:
    """
    Base class of a graph-like ADT.

    This wraps a igraph.Graph, but with several differences:
    * uses a unique string key to access the nodes that will *NOT* be affected by the removal of other nodes;
    * checks invariants to ensure the validity of the data structure when modifying the nodes and edges;
    * all modification operations called on nodes and edges will also be routed back to the container class for checking the invariants.

    "Modification" means adding or removing nodes and edges. It does not include changing the properties of nodes or edges.
    """

    def __init__(self, directed: bool = True):
        self.g: igraph.Graph = igraph.Graph(directed=directed)
        self.next_id: int = 0

    @abc.abstractmethod
    def check_invariants(self) -> bool:
        """
        Checks if the data structure is in a valid state.
        Default implementation checks nothing, which allows any graph. Subclasses should override this method to implement the specific invariants.
        """
        return True

    @abc.abstractmethod
    def check_invariants_before_adding_node(
        self,
        parents: Optional[List[Node]] = None,
        children: Optional[List[Node]] = None,
    ) -> bool:
        """
        Checks if adding a node with the given parents and children is valid, *before* adding the node.
        Default implementation does nothing here and defers the invariants checking to after adding the node. Subclasses can override this method to perform incremental checks of invariants.
        """
        return True

    @abc.abstractmethod
    def check_invariants_after_adding_node(
        self,
        node: Node,
        parents: Optional[List[Node]] = None,
        children: Optional[List[Node]] = None,
    ) -> bool:
        """
        Checks if adding a node with the given parents and children is valid, *after* adding the node.
        Default implementation calls check_invariants to check the entire data structure. Subclasses can override this method to perform incremental checks of invariants, either here or before adding the node.
        """
        return self.check_invariants()

    def add_node(
        self,
        parents: Optional[List[Node]] = None,
        children: Optional[List[Node]] = None,
    ) -> Node:
        if not self.check_invariants_before_adding_node(parents, children):
            raise InvariantError("Adding a node failed due to invariant violation.")
        node = self._add_node()
        if parents is not None:
            for other in parents:
                self._add_edge(other.key, node.key)
        if children is not None:
            for other in children:
                self._add_edge(node.key, other.key)
        if not self.check_invariants_after_adding_node(node, parents, children):
            raise InvariantError("Adding a node failed due to invariant violation.")
        return node

    # TODO: delete_node, add_edge, and delete_edge

    def _add_node(self) -> Node:
        key = f"n{self.next_id}"
        self.next_id += 1
        return Node(self.g.add_vertex(name=key), self)

    def _delete_node(self, key: str) -> None:
        self.g.delete_vertices(key)

    def _add_edge(self, src: str, tgt: str) -> Edge:
        return Edge(self.g.add_edge(src, tgt), self)

    def _delete_edge(self, src: str, tgt: str):
        self.g.delete_edges([(src, tgt)])

    def get_node(self, key: str) -> Node:
        return Node(self.g.vs.find(name=key), self)

    def get_edge(self, from_key: str, to_key: str) -> Edge:
        return Edge(self.g.es.find(from_key=from_key, to_key=to_key), self)

    def ncount(self) -> int:
        return self.g.vcount()

    def ecount(self) -> int:
        return self.g.ecount()

    def __str__(self):
        return self.g.__str__()

    def __repr__(self):
        return self.g.__repr__()

    def nodes(self) -> Iterable[Node]:
        return [Node(v, self) for v in self.g.vs]

    def edges(self) -> Iterable[Edge]:
        return [Edge(e, self) for e in self.g.es]



import signal
from contextlib import contextmanager


class TimeoutException(Exception):
    pass


class TimeUtils:
    @classmethod
    @contextmanager
    def time_limit(cls, seconds):
        def signal_handler(signum, frame):
            raise TimeoutException("Timed out after {} seconds!".format(seconds))
        signal.signal(signal.SIGALRM, signal_handler)
        signal.alarm(seconds)
        try:
            yield
        finally:
            signal.alarm(0)



from pathlib import Path
import os
import sys
import yaml
from typing import *


class Macros:
    THIS_DIR: Path = Path(os.path.dirname(os.path.realpath(__file__)))
    SYS_DIR: Path = Path(sys.prefix)
    DEFAULT_CONFIG_FILE: Path = SYS_DIR / "config" / "seutil_config.json"

    if os.getenv("HOME") is not None:
        HOME_DIR: Path = Path.home()
        CONFIG_FILE: Path = HOME_DIR / ".seutil"
    else:
        CONFIG_FILE: Path = DEFAULT_CONFIG_FILE
    # end if

def get_config(key: str,
               config_file: Path = Macros.CONFIG_FILE,
               is_use_default: bool = True) -> any:
    try:
        with open(str(config_file), "r") as f:
            configs = yaml.load(f, Loader=yaml.FullLoader)
            return configs[key]
        # end with
    except (KeyError, FileNotFoundError) as e:
        if is_use_default and config_file != Macros.DEFAULT_CONFIG_FILE:
            return get_config(key, Macros.DEFAULT_CONFIG_FILE, False)
        else:
            raise e
        # end if
    # end try



from typing import *

from collections import defaultdict
from enum import Enum
import inspect
import json
import os
from pathlib import Path
import pickle as pkl
import pydoc
import recordclass
import shutil
import subprocess
import typing_inspect
import yaml


def is_obj_record_class(obj: Any) -> bool:
    return obj is not None \
           and isinstance(obj, recordclass.mutabletuple) or isinstance(obj, recordclass.dataobject)


def is_clz_record_class(clz: Type) -> bool:
    return clz is not None \
           and inspect.isclass(clz) \
           and (issubclass(clz, recordclass.mutabletuple) or issubclass(clz, recordclass.dataobject))



class IOUtils:
    """
    Utility functions for I/O.
    """

    # ----------
    # Directory operations

    class cd:
        """
        Change directory. Usage:

        with IOUtils.cd(path):
            <statements>
        # end with

        Using a string path is supported for backward compatibility.
        Using pathlib.Path should be preferred.
        """

        def __init__(self, path: Union[str, Path]):
            if isinstance(path, str):
                path = Path(path)
            # end if
            self.path = path  # Path
            self.old_path = Path.cwd()  # Path
            return

        def __enter__(self):
            os.chdir(self.path)
            return

        def __exit__(self, type, value, tb):
            os.chdir(self.old_path)
            return

    # Deprecated
    # Use pathlib.Path.is_dir() instead
    @classmethod
    def has_dir(cls, dirname) -> bool:
        return os.path.isdir(dirname)

    # Deprecated
    # Use pathlib.Path.mkdir() instead
    @classmethod
    def mk_dir(cls, dirname, mode=0o777,
               is_remove_if_exists: bool = False,
               is_make_parent: bool = True):
        """
        Makes the directory.
        :param dirname: the name of the directory.
        :param mode: mode of the directory.
        :param is_remove_if_exists: if the directory with name already exists, whether to remove.
        :param is_make_parent: if make parent directory if not exists.
        """
        if cls.has_dir(dirname):
            if is_remove_if_exists:
                rm_cmd = "rm {} -rf".format(dirname)
                subprocess.run(["bash", "-c", rm_cmd])
            else:
                return
        # end if
        parent_dir = os.path.dirname(dirname)
        if not cls.has_dir(parent_dir):
            if is_make_parent:
                cls.mk_dir(parent_dir, mode, is_remove_if_exists=False, is_make_parent=True)
            else:
                raise FileNotFoundError("Path not found: {}".format(parent_dir))
        # end if
        os.mkdir(dirname, mode)
        return

    @classmethod
    def rm_dir(
            cls,
            path: Path,
            ignore_non_exist: bool = True,
            force: bool = True,
    ):
        """
        Removes the directory.
        :param path: the name of the directory.
        :param ignore_non_exist: ignores error if the directory does not exist.
        :param force: force remove the directory even it's non-empty.
        """
        if path.is_dir():
            if force:
                shutil.rmtree(path, ignore_errors=True)
            else:
                path.rmdir()
            # end if
        else:
            if ignore_non_exist:
                return
            else:
                raise FileNotFoundError("Trying to remove non-exist directory {}".format(path))
            # end if
        # end if
        return

    @classmethod
    def rm(
            cls,
            path: Path,
            ignore_non_exist: bool = True,
            force: bool = True,
    ):
        """
        Removes the file/dir.
        :param path: the path to the file/dir to remove.
        :param ignore_non_exist: ignores error if the file/dir does not exist.
        :param force: force remove the file even it's protected / dir even it's non-empty.
        """
        if path.is_dir():
            cls.rm_dir(path, ignore_non_exist=ignore_non_exist, force=force)
        else:
            if path.exists():
                path.unlink()
            else:
                if not ignore_non_exist:
                    raise IOError(f"{path} does not exist")

    # ----------
    # File operations

    class Format(Enum):
        txt = 0,  # Plain text format
        pkl = 1,  # Pickle format
        jsonPretty = 2,  # Json format, with pretty-printing
        jsonNoSort = 3,  # Json format, with pretty-printing, without sorting the keys in dictionary
        json = 4,  # Json format, without pretty-printing (eveything on one line)
        jsonList = 5,  # Json format, assuming a list structure and put each item on one line
        txtList = 6,  # Plain text format, dump/load as a list where each line is an element
        yaml = 7,  # YAML format

        @classmethod
        def from_str(cls, string: str) -> "IOUtils.Format":
            return {
                "pkl": IOUtils.Format.pkl,
                "json": IOUtils.Format.jsonPretty,
                "json-nosort": IOUtils.Format.jsonNoSort,
                "json_nosort": IOUtils.Format.jsonNoSort,
                "json-min": IOUtils.Format.json,
                "json_min": IOUtils.Format.json,
            }.get(string, IOUtils.Format.txt)

        def get_extension(self) -> str:
            return {
                IOUtils.Format.txt: "txt",
                IOUtils.Format.pkl: "pkl",
                IOUtils.Format.jsonPretty: "json",
                IOUtils.Format.jsonNoSort: "json",
                IOUtils.Format.json: "json",
                IOUtils.Format.jsonList: "jsonl",
                IOUtils.Format.txtList: "txt",
                IOUtils.Format.yaml: "yml",
            }.get(self, "unknown")

    IO_FORMATS: Dict[Format, Dict] = defaultdict(lambda: {
        "mode": "t",
        "dumpf": (lambda obj, f: f.write(obj)),
        "loadf": (lambda f: f.read())
    })

    IO_FORMATS[Format.pkl]["mode"] = "b"
    IO_FORMATS[Format.pkl]["dumpf"] = lambda obj, f: pkl.dump(obj, f, protocol=pkl.HIGHEST_PROTOCOL)
    IO_FORMATS[Format.pkl]["loadf"] = lambda f: pkl.load(f)

    IO_FORMATS[Format.jsonPretty]["dumpf"] = lambda obj, f: json.dump(obj, f, indent=4, sort_keys=True)
    IO_FORMATS[Format.jsonPretty]["loadf"] = lambda f: yaml.load(f, Loader=yaml.FullLoader)  # allows some format errors (e.g., trailing commas)

    IO_FORMATS[Format.jsonNoSort]["dumpf"] = lambda obj, f: json.dump(obj, f, indent=4)
    IO_FORMATS[Format.jsonNoSort]["loadf"] = lambda f: yaml.load(f, Loader=yaml.FullLoader)  # allows some format errors (e.g., trailing commas)

    IO_FORMATS[Format.json]["dumpf"] = lambda obj, f: json.dump(obj, f, sort_keys=True)
    IO_FORMATS[Format.json]["loadf"] = lambda f: json.load(f)

    IO_FORMATS[Format.yaml]["dumpf"] = lambda obj, f: yaml.dump(obj, f)
    IO_FORMATS[Format.yaml]["loadf"] = lambda f: yaml.load(f, Loader=yaml.FullLoader)

    @classmethod
    def dumpf_json_list(cls, obj, f):
        for item in obj:
            f.write(json.dumps(item) + "\n")

    @classmethod
    def loadf_json_list(cls, f) -> List:
        obj = []
        for line in f.readlines():
            obj.append(json.loads(line))
        return obj

    IO_FORMATS[Format.jsonList]["dumpf"] = lambda obj, f: IOUtils.dumpf_json_list(obj, f)
    IO_FORMATS[Format.jsonList]["loadf"] = lambda f: IOUtils.loadf_json_list(f)

    @classmethod
    def dumpf_txt_list(cls, obj, f):
        for item in obj:
            f.write(str(item) + "\n")

    @classmethod
    def loadf_txt_list(cls, f) -> List:
        return f.read().splitlines()

    IO_FORMATS[Format.txtList]["dumpf"] = lambda obj, f: IOUtils.dumpf_txt_list(obj, f)
    IO_FORMATS[Format.txtList]["loadf"] = lambda f: IOUtils.loadf_txt_list(f)

    @classmethod
    def dump(
            cls,
            file_path: Union[str, Path],
            obj: object,
            fmt: Union[Format, str] = Format.jsonPretty,
            append: bool = False,
    ) -> None:
        """
        Saves an object to the file in the specified format.
        By default, the format is json pretty-print, and the existing content in the file will be erased.
        :param file_path: the file to save the object into.
        :param obj: the object to save.
        :param fmt: the format, one of IOUtils.Format.
        :param append: if true, appends to the file instead of erasing existing content in the file.
        """
        if isinstance(file_path, str):
            file_path = Path(file_path)

        file_path.touch(exist_ok=True)

        if isinstance(fmt, str):  fmt = cls.Format.from_str(fmt)
        conf = cls.IO_FORMATS[fmt]

        write_mode = "w" if not append else "a"
        with open(file_path, write_mode + conf["mode"]) as f:
            conf["dumpf"](obj, f)

        return

    @classmethod
    def load(cls, file_path: Union[str, Path], fmt: Union[Format, str] = Format.jsonPretty) -> Any:
        if isinstance(file_path, str):
            file_path = Path(file_path)
        # end if

        if isinstance(fmt, str):  fmt = cls.Format.from_str(fmt)
        conf = cls.IO_FORMATS[fmt]

        try:
            with open(file_path, "r" + conf["mode"]) as f:
                obj = conf["loadf"](f)
            # end with
        except FileNotFoundError as e:
            raise FileNotFoundError(str(e) + " at {}".format(Path.cwd()))
        # end try

        return obj

    @classmethod
    def load_json_stream(cls, file_path: Union[str, Path], fmt: Union[Format, str] = Format.jsonPretty):
        """
        Reads large json file containing a list of data iteratively. Returns a generator function.
        """
        import ijson
        if isinstance(file_path, str):
            file_path = Path(file_path)
        # end if

        if isinstance(fmt, str):  fmt = cls.Format.from_str(fmt)
        conf = cls.IO_FORMATS[fmt]

        try:
            with open(file_path, "r" + conf["mode"]) as f:
                objects = ijson.items(f, "item")
                for obj in objects:
                    yield obj
            # end with
        except FileNotFoundError as e:
            raise FileNotFoundError(str(e) + " at {}".format(Path.cwd()))
        # end try

    @classmethod
    def update_json(cls, file_name, data):
        """
        Updates the json data file. The data should be dict like (support update).
        """
        try:
            orig_data = cls.load(file_name)
        except:
            orig_data = dict()
        # end try
        orig_data.update(data)
        cls.dump(file_name, orig_data)
        return orig_data

    @classmethod
    def extend_json(cls, file_name, data):
        """
        Updates the json data file. The data should be list like (support extend).
        """
        try:
            orig_data = cls.load(file_name)
        except:
            orig_data = list()
        # end try
        orig_data.extend(data)
        cls.dump(file_name, orig_data)
        return orig_data

    JSONFY_FUNC_NAME = "jsonfy"
    DEJSONFY_FUNC_NAME = "dejsonfy"
    JSONFY_ATTR_FIELD_NAME = "jsonfy_attr"

    @classmethod
    def jsonfy(cls, obj):
        """
        Turns an object to a json-compatible data structure.
        A json-compatible data can only have list, dict (with str keys), str, int and float.
        Any object of other classes will be casted through (try each option in order, if applicable):
        1. JSONFY function, which takes no argument and returns a json-compatible data;
           should have the name {@link IOUtils#JSONFY_FUNC_NAME};
        2. JSONFY_ATTR field, which is a dict of attribute name-type pairs, that will be extracted from the object to a dict;
           should have the name {@link IOUtils#JSONFY_ATTR_FIELD_NAME};
        3. cast to a string.
        """
        if obj is None:
            return None
        elif isinstance(obj, (int, float, str, bool)):
            # primitive types
            return obj
        elif isinstance(obj, (list, set, tuple)):
            # array
            return [cls.jsonfy(item) for item in obj]
        elif isinstance(obj, dict):
            # dict
            return {k: cls.jsonfy(v) for k, v in obj.items()}
        elif isinstance(obj, Enum):
            # Enum
            return obj.value
        elif hasattr(obj, cls.JSONFY_FUNC_NAME):
            # with jsonfy function
            return getattr(obj, cls.JSONFY_FUNC_NAME)()
        elif hasattr(obj, cls.JSONFY_ATTR_FIELD_NAME):
            # with jsonfy_attr annotations
            return {attr: cls.jsonfy(getattr(obj, attr)) for attr in getattr(obj, cls.JSONFY_ATTR_FIELD_NAME).keys()}
        elif is_obj_record_class(obj):
            # RecordClass
            if hasattr(obj, "__dict__"):
                # Older versions of recordclass
                return {k: cls.jsonfy(v) for k, v in obj.__dict__.items()}
            else:
                # Newer versions of recordclass
                return {k: cls.jsonfy(getattr(obj, k)) for k in obj.__fields__}
        else:
            # Last effort: toString
            return repr(obj)

    @classmethod
    def dejsonfy(cls, data, clz: Optional[Union[Type, str]] = None):
        """
        Turns a json-compatible data structure to an object of class {@code clz}.
        If {@code clz} is not assigned, the data will be casted to dict or list if possible.
        Otherwise the data will be casted to the object through (try each option in order, if applicable):
        1. DEJSONFY function, which takes the data as argument and returns a object;
           should have the name {@link IOUtils#DEJSONFY_FUNC_NAME};
        2. JSONFY_ATTR field, which is a dict of attribute name-type pairs, that will be extracted from the object to a dict;
           should have the name {@link IOUtils#JSONFY_ATTR_FIELD_NAME};
        """
        if isinstance(clz, str):
            clz = pydoc.locate(clz)

        if data is None:
            # None value
            return None
        elif clz is not None and typing_inspect.get_origin(clz) == list:
            # List[XXX]
            return [cls.dejsonfy(item, clz.__args__[0]) for item in data]
        elif clz is not None and typing_inspect.get_origin(clz) == tuple:
            # Tuple[XXX]
            return tuple([cls.dejsonfy(item, clz.__args__[min(i, len(clz.__args__)-1)]) for i, item in enumerate(data)])
        elif clz is not None and typing_inspect.get_origin(clz) == set:
            # Set[XXX]
            return set([cls.dejsonfy(item, clz.__args__[0]) for item in data])
        elif clz is not None and hasattr(clz, cls.DEJSONFY_FUNC_NAME):
            # with dejsonfy function
            return clz.dejsonfy(data)
        elif isinstance(data, list):
            # array
            return [cls.dejsonfy(item, clz) for item in data]
        elif clz is not None and hasattr(clz, cls.JSONFY_ATTR_FIELD_NAME):
            # with jsonfy_attr annotations
            obj = clz()
            for attr, attr_clz in getattr(clz, cls.JSONFY_ATTR_FIELD_NAME).items():
                if attr in data:
                    setattr(obj, attr, cls.dejsonfy(data[attr], attr_clz))
            return obj
        elif clz is not None and is_clz_record_class(clz):
            # RecordClass
            field_values = dict()
            for f, t in get_type_hints(clz).items():
                if f in data:
                    field_values[f] = cls.dejsonfy(data.get(f), t)
            return clz(**field_values)
        elif clz is not None and inspect.isclass(clz) and issubclass(clz, Enum):
            # Enum
            return clz(data)
        elif isinstance(data, dict):
            # dict
            return {k: cls.dejsonfy(v, clz) for k, v in data.items()}
        else:
            # primitive types / unresolvable things
            if clz is not None:
                try:
                    return clz(data)
                except:
                    pass
            return data



import argparse
from typing import *

from .LoggingUtils import LoggingUtils


class Option:
    def __init__(self):
        self.name = str()
        self.type = "str"
        self.str_value = str()


class StoreInDict(argparse.Action):
    # def parse_values(self, values) -> Tuple[Sequence[Option], Sequence[str]]:
    #     i = 0
    #     length = len(values)
    #
    #     args = list()
    #     options = list()
    #     new_option = None
    #
    #     while i < length:
    #         cur_item = values[i]
    #         if cur_item.startswith("-"):
    #             if new_option is not None:
    #                 options.append(new_option)
    #             # end if
    #             new_option = Option()
    #             # Value
    #             if "=" in cur_item:
    #                 new_option.
    #     # end while
    #     if new_option is not None:
    #         options.append(new_option)
    #     # end if
    #     return options, args


    def __call__(self, parser, namespace, values, option_string=None):
        d = getattr(namespace, self.dest)
        types = dict()
        for opt in values:
            try:
                name, value = opt.split("=", 1)
            except:
                name = opt
                value = True
            # end try
            name = name.lstrip("-")
            if ":" in name:
                name, types[name] = name.split(":", 1)
            else:
                types[name] = ""
            if name in d:
                d[name].append(value)
            else:
                d[name] = [value]
        for name in d:
            if not types[name] == "list" and len(d[name]) == 1:
                d[name] = d[name][0]
            # end if
            if types[name] != "":
                d[name] = eval("{}(\"{}\")".format(types[name], d[name]))
                continue
            # end if
            try:
                intval = int(d[name])
            except:
                pass
            else:
                d[name] = intval
                continue
            # end try
            try:
                floatval = float(d[name])
            except:
                pass
            else:
                d[name] = floatval
                continue
            # end try
        setattr(namespace, self.dest, d)


def main(argv, actions: Dict[str, Callable], normalize_options: Callable[[Dict], Dict] = None):
    """
    Main function for command line option parsing, in the form of "THIS action options...",
    Where each option is in the form of "-name=value".
    :param argv: The command line inputs, without the name of the script (sys.argv[1:]).
    :param actions: The mapping from action name to action function.
    :param normalize_options: Optional function to normalize options, by default identical function.
    """
    logger = LoggingUtils.get_logger("CliUtils.main")
    logger.info("Starting")

    if normalize_options is None:
        normalize_options = lambda opts: opts

    if len(argv) == 0 or argv[0].startswith("-"):
        action = "default_action"
        cli_options = argv
    else:
        action = argv[0]
        cli_options = argv[1:]
    # end if

    # Prevent distinguish between positional arguments and optional arguments (HACK)
    p = argparse.ArgumentParser(prefix_chars=' ')
    p.add_argument("options", nargs="*", action=StoreInDict, default=dict())
    options = p.parse_args(cli_options).options

    # normalize options
    options = normalize_options(options)

    if action in actions:
        actions[action](**options)
    else:
        print("No such action {}".format(action))
        print("Available actions: {}".format(list(actions.keys())))
    # end if

    logger.info("Terminating")
    return



import pytest
import seutil as su


def test_lattice_add():
    lattice = su.ds.lattice.Lattice()
    n0 = lattice.add_node()
    n1 = lattice.add_node(parents=[n0])
    n2 = lattice.add_node(parents=[n0])
    n3 = lattice.add_node(parents=[n1, n2])
    assert lattice.ncount() == 4
    assert lattice.ecount() == 4


def test_lattice_acyclic():
    lattice = su.ds.lattice.Lattice()
    n0 = lattice.add_node()
    n1 = lattice.add_node(parents=[n0])
    n2 = lattice.add_node(parents=[n0])
    with pytest.raises(su.ds.InvariantError):
        n3 = lattice.add_node(parents=[n1, n2], children=[n0])


def test_lattice_connected():
    lattice = su.ds.lattice.Lattice(connected=True)
    n0 = lattice.add_node()
    with pytest.raises(su.ds.InvariantError):
        n1 = lattice.add_node()


def test_lattice_connected_is_by_default_off():
    lattice = su.ds.lattice.Lattice()
    n0 = lattice.add_node()
    n1 = lattice.add_node()
    assert lattice.ncount() == 2
    assert lattice.ecount() == 0


def test_lattice_entry_exit_nodes():
    lattice = su.ds.lattice.Lattice()
    n0 = lattice.add_node()
    n1 = lattice.add_node(parents=[n0])
    n2 = lattice.add_node(parents=[n0])
    n3 = lattice.add_node(parents=[n1, n2])
    assert lattice.entry_nodes == {n0}
    assert lattice.exit_nodes == {n3}



from typing import *

import os
from pathlib import Path
import subprocess


class BashUtils:
    """
    Utility functions for running Bash commands.
    """
    PRINT_LIMIT = 1000

    class RunResult(NamedTuple):
        return_code: int
        stdout: str
        stderr: str

    @classmethod
    def run(cls, cmd: str,
            expected_return_code: int = None,
            is_update_env: bool = False,
            timeout: Optional[float] = None,
    ) -> RunResult:
        """
        Runs a Bash command and returns the stdout.
        :param cmd: the command to run.
        :param expected_return_code: if set to an int, will raise exception if the return code mismatch.
        :param is_update_env: if true, the environment in *this python process (os.environ)* will be updated upon the successful execution of cmd (i.e., returns 0), to reflect the changes to the enrionment variables cmd may make.  Note it can not change the environment of the process that invoked this python process.  It is useful because the updated environment will be used for later BashUtils.run executions.
        :param timeout: if not None, kill the process after timeout seconds and raise TimeoutExpire exception.
        :return: the run result, which is a named tuple with field return_code, stdout, stderr.
        """
        # If update env is requested, append an additional command to the cmd
        if is_update_env:
            tempfile_update_env = cls.get_temp_file()
            cmd += f" && env > {tempfile_update_env}"
        # end if

        completed_process = subprocess.run(["bash", "-c", cmd], stdout=subprocess.PIPE, stderr=subprocess.PIPE, timeout=timeout)
        #completed_process = subprocess.run(cmd, shell=True, executable="/bin/bash", stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        return_code = completed_process.returncode
        stdout = completed_process.stdout.decode("utf-8", errors="ignore")
        stderr = completed_process.stderr.decode("utf-8", errors="ignore")

        # Update env, if requested and return code is 0
        if is_update_env and return_code == 0:
            with open(str(tempfile_update_env), "r") as fp:
                os.environ.clear()
                for line in fp.read().splitlines(keepends=False):
                    env_key, env_value = line.split(sep="=", maxsplit=1)
                    os.environ[env_key] = env_value
                # end for
            # end with
        # end if

        if expected_return_code is not None:
            if return_code != expected_return_code:
                if len(stdout) > cls.PRINT_LIMIT:
                    tempfile_stdout = subprocess.run(["bash", "-c", "mktemp"], stdout=subprocess.PIPE, stderr=subprocess.PIPE).stdout.decode("utf-8", errors="ignore").strip()
                    with open(tempfile_stdout, "w") as f:
                        f.write(stdout)
                    # end with
                    stdout = f"{stdout[:cls.PRINT_LIMIT]} //////////TOO LONG; dumped to {tempfile_stdout}//////////"
                # end if
                if len(stderr) > cls.PRINT_LIMIT:
                    tempfile_stderr = subprocess.run(["bash", "-c", "mktemp"], stdout=subprocess.PIPE, stderr=subprocess.PIPE).stdout.decode("utf-8", errors="ignore").strip()
                    with open(tempfile_stderr, "w") as f:
                        f.write(stderr)
                    # end with
                    stderr = f"{stderr[:cls.PRINT_LIMIT]} //////////TOO LONG; dumped to {tempfile_stderr}//////////"
                # end if
                raise RuntimeError(f"Expected {expected_return_code} but returned {return_code} while executing bash command '{cmd}'.\nstdout: {stdout}\nstderr: {stderr}")
        # end if, if

        return cls.RunResult(return_code, stdout, stderr)

    @classmethod
    def get_temp_dir(cls) -> Path:
        return Path(cls.run("mktemp -d", expected_return_code=0).stdout.strip())

    @classmethod
    def get_temp_file(cls) -> Path:
        return Path(cls.run("mktemp", expected_return_code=0).stdout.strip())



import collections
from typing import Any, Dict, List, Optional

from varname import argname


class Reporter:
    def __init__(self):
        self.history: Dict[str, List[Any]] = collections.defaultdict(list)

    def add_to_history(self, name, value):
        self.history[name].append(value)

    def generate_report(self) -> str:
        s = "===== Report =====\n"
        s += f"Inspected {len(self.history)} variables with {sum(len(x) for x in self.history.values())} values\n"
        for name, values in self.history.items():
            s += f"--- {name}\n"
            s += f"count: {len(values)}; unique count: {len(set(values))};\n"
        s += "==================\n"
        return s


default_reporter = Reporter()


def inspect(
    var: Any,
    name: Optional[str] = None,
    reporter: Reporter = default_reporter,
):
    if name is None:
        name = argname("var", vars_only=False)
    reporter.add_to_history(name, var)


def report(
    reporter: Reporter = default_reporter,
):
    print(reporter.generate_report())



from pathlib import Path
import numpy as np
import random
import subprocess
from typing import *

from .IOUtils import IOUtils


class Stream:
    """
    Streams help manipulate sequences of objects.
    """
    def __init__(self):
        self.items = list()
        return

    @classmethod
    def of(cls, one_or_more_items):
        """
        Get a new stream from the item / items.
        :param one_or_more_items: is converted to list with builtin `list` function.
        """
        stream = Stream()
        if one_or_more_items is not None:
            stream.items = list(one_or_more_items)
        # end if, if
        return stream

    @classmethod
    def of_files(cls, dir_path: Union[str, Path]):
        """
        Get a stream of the files under the directory.
        """
        with IOUtils.cd(dir_path):
            cmd_find = "find -mindepth 1 -maxdepth 1 -type f"
            files = subprocess.run(["bash","-c",cmd_find], stdout=subprocess.PIPE).stdout.decode("utf-8").split("\n")[:-1]
        # end with
        files = [file[2:] for file in files]
        stream = cls.of(files)
        stream.sorted()
        return stream

    @classmethod
    def of_dirs(cls, dir_path: Union[str, Path]):
        """
        Get a stream of the sub-directories under the directory.
        """
        with IOUtils.cd(dir_path):
            cmd_find = "find -mindepth 1 -maxdepth 1 -type d"
            dirs = subprocess.run(["bash","-c",cmd_find], stdout=subprocess.PIPE).stdout.decode("utf-8").split("\n")[:-1]
        # end with
        dirs = [dir[2:] for dir in dirs]
        stream = cls.of(dirs)
        stream.sorted()
        return stream

    def filter(self, predicate_func: Callable[[object], bool]):
        """
        Returns a stream consisting of the elements of this stream that match the given predicate.
        """
        return Stream.of(item for item in self.items if predicate_func(item))

    def count(self):
        return sum(self.items)

    def reduce(self, count_func: Callable[[str], float] = lambda x: 1):
        return sum([count_func(f) for f in self.items])

    def sorted(self, key: Callable[[str], object] = lambda f: f,
               reverse: bool = False):
        """
        Sorts the list of files in the dataset.
        """
        list.sort(self.items, key=key, reverse=reverse)
        return self

    def map(self, map_func: Callable[[str], object],
            errors: str = "raise", default: object = ""):
        def new_items_generator():
            for item in self.items:
                try:
                    new_item = map_func(item)
                except:
                    if errors == "ignore":
                        yield default
                    else:
                        raise
                else:
                    yield new_item
            # end for
        # end def
        return Stream.of(new_items_generator())

    def peak(self, peak_func: Callable[[str], None],
             errors: str = "ignore"):
        for item in self.items:
            try:
                peak_func(item)
            except:
                if errors == "ignore":
                    continue
                else:
                    raise
        # end for
        return self

    def split(self, fraction_list: List[float],
              count_func: Callable[[str], float] = lambda x: 1):
        """
        Splits the dataset as each part specified by the fractions (assumed to sum up to 1).
        Splitting is done by finding the cutting points. If randomization is needed, call shuffle first.
        :param count_func: customize the number of data counts in each file.
        """
        if self.is_empty():
            return tuple(Stream() for i in range(len(fraction_list)))

        count_list = [count_func(f) for f in self.items]
        cum_count_list = np.cumsum(count_list)
        cum_expected_count_list = [f * cum_count_list[-1] for f in np.cumsum(fraction_list)]
        cut_index_list = []
        last_i = 0
        for i, cum_count in enumerate(cum_count_list):
            if cum_count >= cum_expected_count_list[len(cut_index_list)]:
                last_i = i+1
                cut_index_list.append(i+1)
                if len(cut_index_list) >= len(cum_expected_count_list):
                    break
                # end if
        # end for if
        if last_i != len(cum_count_list):
            cut_index_list.append(len(cum_count_list))
        # end if
        cut_index_list.insert(0,0)
        return tuple(Stream.of(self.items[cut_index_list[i]:cut_index_list[i + 1]]) for i in range(len(cut_index_list) - 1))

    def shuffle(self, seed=None):
        """
        Shuffles the list of files in the dataset.
        """
        random.seed(seed)
        random.shuffle(self.items)
        return self

    def get(self, index: int):
        return self.items[index]

    def is_empty(self):
        return len(self.items) == 0

    def __getitem__(self, item):
        new_items = self.items.__getitem__(item)
        if not isinstance(item, slice):
            new_items = [new_items]
        return Stream.of(new_items)

    def __setitem__(self, key, value):
        return self.items.__setitem__(key, value)

    def __delitem__(self, key):
        return self.items.__delitem__(key)

    def __iter__(self):
        return self.items.__iter__()

    def __len__(self):
        return self.items.__len__()

    def __str__(self):
        return "Stream with {} items".format(len(self.items))

    def __repr__(self):
        return self.__str__()

    def __add__(self, other):
        if isinstance(other, Stream):
            return Stream.of(self.items+other.items)
        else:
            raise NotImplementedError



import math
import traceback
from datetime import datetime
from time import sleep
from typing import *

from github import Github, RateLimitExceededException
from github.GithubException import GithubException
from github.NamedUser import NamedUser
from github.Repository import Repository

from . import _config
from .LoggingUtils import LoggingUtils
from .BashUtils import BashUtils


class GitHubUtils:

    logger = LoggingUtils.get_logger("GitHubUtils", LoggingUtils.DEBUG)

    GITHUB_SEARCH_ITEMS_MAX = 1000
    try:
        DEFAULT_ACCESS_TOKEN = _config.get_config("github_access_token")
        DEFAULT_GITHUB_OBJECT = Github(DEFAULT_ACCESS_TOKEN, per_page=100)
    except:
        DEFAULT_ACCESS_TOKEN = None
        DEFAULT_GITHUB_OBJECT = None
        logger.info("Fail to get github_access_token from config file.  Using GitHubUtils APIs will require compulsory input access_token")
    # end try

    @classmethod
    def get_github(cls, access_token: str = None) -> Github:
        if access_token is None:
            return cls.DEFAULT_GITHUB_OBJECT
        else:
            return Github(access_token)

    class wait_rate_limit:
        """
        Wait for rate limit of the github accessor. For use with "with".
        Use the default github accessor if no argument is given.
        """
        DEFAULT_GITHUB_OBJECT = None
        logger = None

        def __init__(self, github: Github = DEFAULT_GITHUB_OBJECT):
            self.github = github
            return

        def __enter__(self):
            if self.github is None:
                self.github = self.DEFAULT_GITHUB_OBJECT
            # end if

            # Check rate limit
            rate_limit_remain, rate_limit = self.github.rate_limiting
            if rate_limit_remain <= 1:
                self.logger.debug("Rate limit {} / {}".format(rate_limit_remain, rate_limit))
                rate_limit_reset_time = datetime.fromtimestamp(self.github.rate_limiting_resettime)
                rate_limit_wait_seconds = math.ceil((rate_limit_reset_time - datetime.now()).total_seconds()) + 1
                if rate_limit_wait_seconds > 0:
                    self.logger.warning("Rate limit will recover at: {}, will wait for {} seconds.".format(rate_limit_reset_time, rate_limit_wait_seconds))
                    sleep(rate_limit_wait_seconds)
                    self.logger.warning("Rate limit recovered")
                # end if
            # end if
            return self.github

        def __exit__(self, type, value, tb):
            return

    # end class
    wait_rate_limit.DEFAULT_GITHUB_OBJECT = DEFAULT_GITHUB_OBJECT
    wait_rate_limit.logger = logger

    T = TypeVar("T")
    @classmethod
    def ensure_github_api_call(cls, call: Callable[[Github], T], github: Github = DEFAULT_GITHUB_OBJECT, max_retry_times: int = float("inf")) -> T:
        retry_times = 0
        while True:
            try:
                with cls.wait_rate_limit(github) as g:
                    return call(g)
                # end with
            except (GithubException, RateLimitExceededException) as e:
                if e.status == 422:
                    cls.logger.warning("Validation Error. Will not retry.")
                    raise
                else:
                    cls.logger.warning("Unexpected exception during api call: {}".format(traceback.format_exc()))
                    retry_times += 1
                    if retry_times > max_retry_times:
                        cls.logger.warning("Exceeding max retry times {}".format(max_retry_times))
                        raise
                    # end if

                    retry_wait_time = min(retry_times * 30, 600)
                    cls.logger.warning("Will wait {} seconds before retry {}".format(retry_wait_time, retry_times))
                    sleep(retry_wait_time)
            # end try
        # end while

    @classmethod
    def search_repos(cls, q: str = "", sort: str = "stars", order: str = "desc",
                     is_allow_fork: bool = False,
                     max_num_repos: int = GITHUB_SEARCH_ITEMS_MAX,
                     github: Github = DEFAULT_GITHUB_OBJECT,
                     max_retry_times: int = float("inf"),
                     *_, **qualifiers) -> List[Repository]:
        """
        Searches the repos by querying GitHub API v3.
        :return: a list of full names of the repos match the query.
        """
        cls.logger.debug("Search for repos with query {}, sort {}, order {}".format(q, sort, order))
        repos = list()
        num_repos = 0
        repos_iterator = iter(github.search_repositories(q, sort, order, **qualifiers))
        while True:
            try:
                repo = cls.ensure_github_api_call(lambda g: next(repos_iterator), github, max_retry_times)

                # Check fork
                if not is_allow_fork:
                    if repo.fork:
                        continue
                # end if, if

                repos.append(repo)
                num_repos += 1

                # Check number
                if num_repos >= max_num_repos:
                    break
                # end if
            except StopIteration:
                break
            except:
                cls.logger.warning("Unknown exception: {}".format(traceback.format_exc()))
                cls.logger.warning("Returning partial results")
                break
            # end try except
        # end while

        if num_repos < max_num_repos:
            cls.logger.info("Got {}/{} repos".format(num_repos, max_num_repos))
        else:
            cls.logger.info("Got {}/{} repos".format(num_repos, max_num_repos))
        # end if

        return repos

    @classmethod
    def search_users(cls, q: str = "", sort: str = "repositories", order: str = "desc",
                     max_num_users: int = GITHUB_SEARCH_ITEMS_MAX,
                     github: Github = DEFAULT_GITHUB_OBJECT,
                     max_retry_times: int = float("inf"),
                     *_, **qualifiers) -> List[NamedUser]:
        """
        Searches the users by querying GitHub API v3.
        :return: a list of usernames (login) of the users match the query.
        """
        cls.logger.debug("Search for users with query {}, sort {}, order {}".format(q, sort, order))
        users = list()
        num_users = 0
        users_iterator = iter(github.search_users(q, sort, order, **qualifiers))
        while True:
            try:
                user = cls.ensure_github_api_call(lambda g: next(users_iterator), github, max_retry_times)

                users.append(user)
                num_users += 1

                # Check number
                if num_users >= max_num_users:
                    break
                # end if
            except StopIteration:
                break
            except:
                cls.logger.warning("Unknown exception: {}".format(traceback.format_exc()))
                cls.logger.warning("Returning partial results.")
                break
            # end try except
        # end while

        if num_users < max_num_users:
            cls.logger.warning("Got {}/{} users".format(num_users, max_num_users))
        else:
            cls.logger.info("Got {}/{} users".format(num_users, max_num_users))
        # end if

        return users

    @classmethod
    def search_repos_of_language(cls, language: str, max_num_repos: int = float("inf"),
                                 is_allow_fork: bool = False,
                                 max_retry_times: int = float("inf"),
                                 strategies: List[str] = None) -> List[Repository]:
        """
        Searches for all the repos of the language.
        :return: a list of full names of matching repos.
        """
        if strategies is None:
            strategies = ["search_repos", "search_users"]
        # end if

        # Check supported strategies
        supported_strategies = ["search_repos", "search_users", "enum_users"]
        for strategy in strategies:
            assert strategy in supported_strategies, strategy
        # end for

        names_repos = dict()

        try:
            # Strategy 1: search repos (limited to 1000)
            strategy = "search_repos"
            if strategy in strategies:
                cls.logger.info("Using strategy {}".format(strategy))
                new_repos = cls.search_repos("language:{}".format(language), is_allow_fork=is_allow_fork, max_retry_times=max_retry_times, max_num_repos=max_num_repos)
                for repo in new_repos:
                    names_repos[repo.full_name] = repo
                # end for
                cls.logger.warning("Progress {}/{} repos.".format(len(names_repos), max_num_repos))
                if len(names_repos) >= max_num_repos:
                    return list(names_repos.values())
                # end if
            # end if

            # Strategy 2: search users (~37000?)
            strategy = "search_users"
            if strategy in strategies:
                cls.logger.info("Using strategy {}".format(strategy))
                s_users = set()
                # s_users = s_users.union([u.login for u in cls.search_users("language:{}".format(language), sort="repositories", max_retry_times=max_retry_times)])
                s_users = s_users.union([u.login for u in cls.search_users("language:{}".format(language), sort="followers", max_retry_times=max_retry_times)])
                # s_users = s_users.union([u.login for u in cls.search_users("language:{}".format(language), sort="joined", max_retry_times=max_retry_times)])
                users_count = 0
                total_users_count = len(s_users)
                for user in s_users:
                    try:
                        new_repos = cls.search_repos("language:{} user:{}".format(language, user), is_allow_fork=is_allow_fork, max_retry_times=max_retry_times)
                    except GithubException as e:
                        cls.logger.warning("Cannot get the repos of user {}".format(user))
                        continue
                    # end try
                    for repo in new_repos:
                        names_repos[repo.full_name] = repo
                    # end for
                    users_count += 1
                    cls.logger.debug("Progress {}/{} repos, {}/{} users.".format(len(names_repos), max_num_repos, users_count, total_users_count))
                    if len(names_repos) >= max_num_repos:
                        return list(names_repos.values())
                    # end if
                # end for
            # end if

            # Strategy 3: enum users (?)
            strategy = "enum_users"
            if strategy in strategies:
                cls.logger.warning("Strategy {} is not implemented yet.".format(strategy))
                cls.logger.warning("Nothing happens.")
            # end if
        except KeyboardInterrupt as e:
            cls.logger.warning("Interrupted. Returning partial results.")
        finally:
            cls.logger.warning("Got {}/{} repos.".format(len(names_repos), max_num_repos))
            return list(names_repos.values())

    @classmethod
    def is_url_valid_git_repo(cls, url: str) -> bool:
        if BashUtils.run(f"git ls-remote {url}").return_code == 0:
            return True
        else:
            return False



import copy
import dataclasses
import re
import sys
import traceback
import warnings
from pathlib import Path
from types import TracebackType
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple

from tqdm import tqdm

from . import bash, ds, io, log

logger = log.get_logger(__name__, log.INFO)


ExcInfo = Tuple[BaseException, BaseException, TracebackType]


class CollatedErrors(Exception):
    def __init__(self, contexts: List[str], exc_infos: List[ExcInfo]):
        assert len(contexts) == len(exc_infos)
        self.contexts = contexts
        self.exc_infos = exc_infos

    def __str__(self):
        s = f"Total {len(self.exc_infos)} errors:"
        for i, (c, e) in enumerate(zip(self.contexts, self.exc_infos)):
            s += f"\n #{i}: {c}\n{traceback.format_exception(*e)}"
        return s

    def __repr__(self):
        return f"{self.__class__.__name__} with {len(self.exc_infos)} errors"


@dataclasses.dataclass
class Project:
    """
    Data structure to represent a software project using a version control system (currently only git is supported).

    A collection of `Project` can be saved to a json/yaml file, as a corpus of software projects.
    `Project` provides common version control operations, such as cloning, checking out, etc.

    :param full_name: the full name of the project; this name is used as the directory/file name when operating on the project, so it should be unique and path-safe (i.e., does not contain '/', ' ', or other special characters that could cause trouble being a path). The recommended name for a project $repo/$name is $repo_$name (it is unique if $repo does not contain '_', which is the case at least for GitHub).
    :param url: the URL to clone the project from.
    :param data: other information of the project stored in key-value pairs, such as revision, branch, etc.; keys should be string and values should be serializable.
    """

    full_name: str
    url: str
    data: Dict[str, Any] = dataclasses.field(default_factory=dict)

    _dir: dataclasses.InitVar[Path] = None

    @property
    def dir(self) -> Path:
        self.require_cloned()
        return self._dir

    @classmethod
    def deserialize(cls, data: dict) -> "Project":
        """
        Deserialize a project from a dictionary.
        The "full_name" and "url" metadata are compulsory and stored in the dataclass;
        the rest of the metadata is stored in the "data" field.
        """
        data = copy.copy(data)
        full_name = data.pop("full_name")
        url = data.pop("url")
        return cls(full_name=full_name, url=url, data=data)

    def serialize(self) -> dict:
        """
        Serialize the metadata of the project (full_name, url, and others in data) to a flat dictionary.
        """
        return {"full_name": self.full_name, "url": self.url, **self.data}

    re_github_url = re.compile(
        r"^((https://)?github\.com/|git@github\.com:)(?P<repo>[^/\n\r]+)/(?P<name>[^/\n\r]+?)(\.git)?$"
    )

    @classmethod
    def from_github_url(cls, url: str) -> "Project":
        """
        Creates a Project from a GitHub URL.
        """
        match = cls.re_github_url.fullmatch(url)
        if match is None:
            raise ValueError(f"Invalid GitHub URL: {url}")
        return cls(full_name=f"{match.group('repo')}_{match.group('name')}", url=url)

    def clone(
        self, downloads_dir: Path, name: Optional[str] = None, exists: str = "ignore"
    ) -> None:
        """
        Clones the project to a local directory, such that more operations can be performed on it.
        :param downloads_dir: the parent directory to clone the project to.
        :param name: the name of the sub-directory to clone the project to; if not provided, `full_name` will be used.
        :param exists: the action to take if the directory already exists:
            * ignore (default): do nothing.
            * remove: remove the directory and clone the project again.
            * pull: update the project by pulling the latest changes; may fail if there is no internet connection or the remote url is not accessible any more.
            * error: raise an error.
        """
        if name is None:
            name = self.full_name
        assert exists in ["ignore", "remove", "pull", "error"]
        logger.info(f"Project {self.full_name}: cloning to {downloads_dir / name}")

        if (downloads_dir / name).exists():
            if exists == "ignore":
                self._dir = downloads_dir / name
                logger.info(
                    f"Project {self.full_name}: existing at {downloads_dir / name}"
                )
                return
            elif exists == "remove":
                io.rmdir(downloads_dir / name)
                logger.info(
                    f"Project {self.full_name}: removed existing at {downloads_dir / name}"
                )
            elif exists == "pull":
                self._dir = downloads_dir / name
                logger.info(
                    f"Project {self.full_name}: existing at {downloads_dir / name}"
                )
                self.fetch()
                return
            elif exists == "error":
                raise RuntimeError(
                    f"Project {self.full_name}: can not clone to existing directory {downloads_dir}/{name}"
                )

        io.mkdir(downloads_dir)
        with io.cd(downloads_dir):
            bash.run(f"git clone {self.url} {name}", 0)
        self._dir = downloads_dir / name

    def set_cloned_dir(self, dir: Path) -> None:
        """
        Tells the project that it has been cloned to the given directory, such that more operations can be performed on it.
        We will trust that you provided a valid directory.
        """
        self._dir = dir
        logger.info(f"Project {self.full_name}: set to be cloned at {dir}")

    def remove(self, error_not_exists: bool = False) -> None:
        """
        Removes the project from the local directory.
        :param error_not_exists: if True, raise an error if the project has not been cloned yet.
        """
        if self._dir is None:
            if error_not_exists:
                raise RuntimeError(
                    f"Project {self.full_name}: not cloned yet, can not remove"
                )
            else:
                logger.info(f"Project {self.full_name}: already removed")
        else:
            io.rmdir(self._dir)
            self._dir = None
            logger.info(f"Project {self.full_name}: removed")

    def require_cloned(self, operation: Optional[str] = None) -> None:
        error_msg = f"Project {self.full_name}: needs to be cloned"
        if operation is not None:
            error_msg += f" before performing the {operation} operation"
        if self._dir is None:
            raise RuntimeError(error_msg)

    def fetch(self) -> None:
        """
        Updates the project by pulling the latest changes.
        May fail if there is no internet connection or the remote url is not accessible any more.
        """
        self.require_cloned("fetch")
        logger.info(f"Project {self.full_name}: fetching latest changes")
        with io.cd(self._dir):
            bash.run("git fetch --all", 0)

    def checkout(self, revision: str, forced: bool = False) -> None:
        """
        Checks out the given revision (or branch or tag) of the project.
        :param revision: the revision (or branch or tag) to checkout.
        :param forced: if True, do force checkout (discarding all local changes that might prevent a checkout).
        """
        self.require_cloned("checkout")
        logger.info(
            f"Project {self.full_name}: checking out revision {revision} ({forced=})"
        )

        with io.cd(self._dir):
            cmd = f"git checkout {revision}"
            if forced:
                cmd += " -f"
            bash.run(cmd, 0)

    def get_cur_revision(self) -> str:
        """
        Gets the current revision of the project.
        """
        self.require_cloned("get_cur_revision")
        with io.cd(self._dir):
            revision = bash.run("git rev-list -1 HEAD", 0).stdout.strip()
        return revision

    def get_revisions_lattice(self) -> ds.lattice.Lattice:
        """
        Gets the revisions lattice of the project, ending at the current revision.
        """
        self.require_cloned("get_revisions_lattice")
        with io.cd(self._dir):
            revisions = bash.run(
                "git rev-list HEAD --topo-order --reverse --parents", 0
            ).stdout.strip()
        revision2node = {}
        lattice = ds.lattice.Lattice()
        for line in revisions.splitlines():
            parts = line.split()
            revision = parts[0]
            node = lattice.add_node(parents=[revision2node[p] for p in parts[1:]])
            revision2node[revision] = node
            node["revision"] = revision
        return lattice

    def for_each_revision(
        self,
        func: Callable[["Project", str], Any],
        revisions: Iterable[str],
        auto_checkout: bool = True,
        errors: str = "collate",
        pbar: Optional[tqdm] = None,
    ) -> List[Any]:
        """
        Runs the func_revision for each revision.

        :param func: the function to run, which takes a Project object and a string revision.
        :param revisions: the revisions to run the function on.
        :param auto_checkout: if set to False, will not automatically checkout each revision (for example, if func does that itself, or if func does not need the project to be checked out at the given revision).
        :param errors: what to do if the function throws errors:
            * ignore: do nothing, not even say a word.
            * warning: make a log.warning for each error.
            * collate (default): collect all errors, and raise a CollatedErrors at the end.
            * error: raise immediately after the first error.
        :param pbar: if provided, the (tqdm) progress bar to use; only the description and the count of the progress bar will be updated, but the total should be set before entering this function.
        :return: the list of return values of the function.
        """
        assert errors in ["ignore", "warning", "collate", "error"]
        self.require_cloned("for_each_revision")

        results = []
        errored_revisions = []
        exc_infos = []

        for revision in revisions:
            if pbar is not None:
                pbar.set_description(f"Revision {revision}")

            try:
                if auto_checkout:
                    self.checkout(revision, True)
                with io.cd(self.dir):
                    results.append(func(self, revision))
            except KeyboardInterrupt:
                raise
            except:
                if errors == "ignore":
                    pass
                elif errors == "warning":
                    warnings.warn(
                        f"Project {self.full_name}: error at revision {revision}: {traceback.format_exc()}"
                    )
                elif errors == "collate":
                    warnings.warn(
                        f"Project {self.full_name}: error at revision {revision}: {traceback.format_exc()}"
                    )
                    errored_revisions.append(revision)
                    exc_infos.append(sys.exc_info())
                elif errors == "error":
                    raise
            finally:
                if pbar is not None:
                    pbar.update(1)

        if errors == "collate" and len(exc_infos) > 0:
            raise CollatedErrors(errored_revisions, exc_infos)
        return results



from random import shuffle
from typing import *

"""
Miscellaneous utility functions.
"""

T = TypeVar("T")


def shuffle_data(items: Sequence[T]) -> Sequence[T]:
    """
    Randomly shuffles the data.
    """
    ran = list(range(len(items)))
    shuffle(ran)
    return [items[i] for i in ran]


def get_num_params(vocab_size, num_layers, num_neurons):
    """
    Returns the number of trainable parameters of an LSTM.

    Args:
        vocab_size (int): The vocabulary size
        num_layers (int): The number of layers in the LSTM
        num_neurons (int): The number of neurons / units per layer

    Returns:
        int: The number of trainable parameters
    """
    num_first_layer = 4 * (num_neurons * (vocab_size + num_neurons) + num_neurons)
    num_other_layer = 4 * (num_neurons * 2 * num_neurons + num_neurons)
    num_softmax = vocab_size * num_neurons + vocab_size

    return num_first_layer + (num_layers - 1) * num_other_layer + num_softmax


def iter_len(iterator: Iterable) -> int:
    """
    Counts the length with the iterator.
    """
    return sum(1 for _ in iterator)


# Human-readable numbers

POWERS = [10 ** x for x in (3, 6, 9, 12, 15, 18, 21, 24)]
HUMAN_READABLE_POWERS = ('K', 'M', 'B', 'T', 'Qa', 'Qi', 'Sx', 'Sp')


def itos_human_readable(value: int, precision: int = 1) -> str:
    """
    Converts a large integer to a human-readable string representation.
    :return the human-readable string representation of the int.
    :raises TypeError if the value passed was unable to be coaxed into int.
    """
    try:
        value = int(value)
    except (TypeError, ValueError) as e:
        raise TypeError("Value can not be converted to int: {}".format(value))

    if value < POWERS[0]:
        return str(value)
    for ordinal, power in enumerate(POWERS[1:], 1):
        if value < power:
            chopped = value / float(POWERS[ordinal - 1])
            fmt = "{0:." + str(precision) + "f}"

            formatted = fmt.format(chopped)
            if "." in formatted:
                formatted = formatted.rstrip("0").rstrip(".")
            # end if
            return formatted + HUMAN_READABLE_POWERS[ordinal - 1]
    return str(value)


def chunks(l, n):
    """Yield successive n-sized chunks from l."""
    for i in range(0, len(l), n):
        yield l[i:i + n]


# Class Property

class ClassPropertyDescriptor(object):

    def __init__(self, fget, fset=None):
        self.fget = fget
        self.fset = fset

    def __get__(self, obj, klass=None):
        if klass is None:
            klass = type(obj)
        return self.fget.__get__(obj, klass)()

    def __set__(self, obj, value):
        if not self.fset:
            raise AttributeError("can't set attribute")
        type_ = type(obj)
        return self.fset.__get__(obj, type_)(value)

    def setter(self, func):
        if not isinstance(func, (classmethod, staticmethod)):
            func = classmethod(func)
        self.fset = func
        return self

def classproperty(func):
    if not isinstance(func, (classmethod, staticmethod)):
        func = classmethod(func)

    return ClassPropertyDescriptor(func)



from pathlib import Path
from typing import *

from .. import IOUtils, LoggingUtils
from .Macro import Macro
from .Table import Table


class File:

    logger = LoggingUtils.get_logger("latex.File")

    def __init__(self, path: Path, is_append: bool = False):
        """
        Creates a latex file, currently for numbers/macros or tables.
        :param path: the path to the file.
        :param is_append: if set to true, then append to any existing file; otherwise overwrite the file.
        """
        self.path: Path = path
        self.is_append: bool = is_append

        self.old_content: str = ""
        if self.is_append:
            if path.is_file():
                self.old_content = IOUtils.load(path, "txt")
        # end if, if

        # Contents
        self.items: List[Union[str, Macro, Table]] = list()
        self.macros_indexed: Dict[str, Macro] = dict()
        return

    def append(self, line: str) -> "File":
        """
        Appends one line of text into the file.
        """
        self.items.append(line)
        return self

    def append_macro(self, macro: Macro) -> "File":
        if macro.key in self.macros_indexed:
            self.logger.warning(f"Redefining macro {macro.key}")
        # end if
        self.macros_indexed[macro.key] = macro
        self.items.append(macro)
        return self

    def append_comment(self, line: str) -> "File":
        self.items.append("%% " + line)
        return self

    def save(self) -> None:
        content = self.old_content + "\n" + self.eval_content()
        IOUtils.dump(self.path, content, "txt")
        return

    def eval_content(self) -> str:
        content_lines: List[str] = list()
        if not self.is_append:
            content_lines.append(self.autogen_notice())
        # end if
        for item in self.items:
            if isinstance(item, str):
                content_lines.append(item)
            elif isinstance(item, Macro):
                content_lines.append(item.eval_content(self.macros_indexed))
            else:
                self.logger.warning("Skipping unsupported item")
            # end if
        # end for
        return "\n".join(content_lines)+"\n"

    def load_macros_from_file(self, file: Path) -> "File":
        self.macros_indexed.update(Macro.load_from_file(file))
        return self

    def __iadd__(self, other: Union[str, Macro]):
        if isinstance(other, str):
            self.append(other)
        elif isinstance(other, Macro):
            self.append_macro(other)
        else:
            self.logger.warning("Skipping unsupported item")
        # end if
        return self

    @classmethod
    def autogen_notice(cls) -> str:
        return "%% Automatically generated by pyutil.latex \n"



import setuptools

with open("README.rst", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="seutil",
    version="0.8.1",
    author="Pengyu Nie",
    author_email="prodigy.sov@gmail.com",
    description="Python utilities for SE research",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/pengyunie/seutil",
    packages=setuptools.find_packages(exclude=["tests"]),
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: POSIX :: Linux",
    ],
    install_requires=[
        "ijson~=3.1.4",
        "igraph>=0.9.10",
        "jsonargparse[all]>=4.1.4,<4.11.0",
        "lazy-import>=0.2.2",
        "numpy>=1.14.4",
        "PyGitHub>=1.40",
        "PyYAML>=5.1",
        "recordclass>=0.11.1",
        "tqdm>=4.62.3,<4.65.0",
        "typing_inspect>=0.4.0",
        "unidiff>=0.5.5",
        "varname>=0.7.1",
    ],
    extras_require={"dev": ["pytest", "flake8", "black", "pytest-cov"]},
)



from pathlib import Path

from jsonargparse.typing import register_type

RPath = Path
RPath.__doc__ = """
Type hint for jsonargparse to obtain a path that is `resolve`d during parsing (relative to cwd if not absolute).
This type does NOT tell jsonargparse to check the existence of the specified path, 
nor does it tell jsonargparse to create the path if it does not exist.
"""


register_type(
    RPath,
    deserializer=lambda v: Path(v).resolve(),
    serializer=lambda v: str(v),
    uniqueness_key=(Path, "ResolvedPath"),
)



import logging
import sys
from logging.handlers import RotatingFileHandler


class LoggingUtils:

    logging_format = "[{relativeCreated:6.0f}{levelname[0]}]{name}: {message}"
    logging_format_detail = "[{asctime}|{relativeCreated:.3f}|{levelname:7}]{name}: {message} [@{filename}:{lineno}|{funcName}|pid {process}|tid {thread}]"

    # Copied from logging
    DEBUG = logging.DEBUG
    INFO = logging.INFO
    WARNING = logging.WARNING
    ERROR = logging.ERROR
    CRITICAL = logging.CRITICAL

    @classmethod
    def get_handler_console(cls, stream=sys.stderr, level=logging.WARNING) -> logging.Handler:
        handler = logging.StreamHandler(stream=stream)
        handler.setLevel(level=level)
        handler.setFormatter(logging.Formatter(cls.logging_format, style="{"))
        return handler

    @classmethod
    def get_handler_file(cls, filename, level=logging.DEBUG) -> logging.Handler:
        handler = RotatingFileHandler(filename, maxBytes=10_000_000, backupCount=1)
        handler.setLevel(level=level)
        handler.setFormatter(logging.Formatter(cls.logging_format_detail, style="{"))
        return handler

    default_level = logging.WARNING
    default_handlers = list()

    @classmethod
    def setup(cls, level=logging.WARNING, filename: str = None):
        logging.basicConfig(level=level, format=cls.logging_format, style="{")

        cls.default_level = level
        cls.default_handlers.clear()
        cls.default_handlers.append(cls.get_handler_console(level=level))
        if filename is not None:
            cls.default_level = logging.DEBUG
            cls.default_handlers.append(cls.get_handler_file(filename=filename))
        # end if
        cls.refresh_loggers()
        return

    loggers = list()

    @classmethod
    def get_logger(cls, name: str,
                   level: int = None) -> logging.Logger:
        if level is None:
            level = cls.default_level
        # end if

        # Always use the same handlers
        handlers = cls.default_handlers

        logger = logging.getLogger(name)
        logger.setLevel(level)
        logger.handlers = []
        logger.propagate = False
        for handler in handlers:
            logger.addHandler(handler)
        # end for
        cls.loggers.append(logger)
        return logger

    @classmethod
    def refresh_loggers(cls):
        """
        Refresh all the loggers to use the default handlers.
        """
        handlers = cls.default_handlers
        for logger in cls.loggers:
            logger.handlers = []
            for handler in handlers:
                logger.addHandler(handler)
            # end for
        # end for
        return

    @classmethod
    def log_and_raise(cls, logger: logging.Logger, msg: str, error_type, level: int = ERROR):
        logger.log(level, msg)
        raise error_type(msg)



import sys
import threading
import time
from typing import Optional, TextIO

from tqdm.std import tqdm as std_tqdm
from tqdm.utils import _screen_shape_wrapper


class PBarManager:
    def __init__(self, out: TextIO, switch_interval: float = 2.5):
        self.out = out
        self.switch_interval = switch_interval

        self.instances = []
        self.cur = 0
        self.last_switch = time.time()

        self.modify_lock = threading.Lock()
        self.cv_update = threading.Condition()
        self.thread = threading.Thread(target=self.run, daemon=True)
        self.thread.start()

    def run(self):
        while True:
            with self.cv_update:
                self.cv_update.wait()

                self.switch()

                self.out.write("\r")
                if self.cur < 0:
                    self.out.flush()
                    continue

                # write summaries of all instances
                summaries = self.format_summaries()
                self.out.write(summaries)

                # write details of the current focused instance
                focus = self.instances[self.cur]
                ncols = _screen_shape_wrapper()(self.out)[0] - len(summaries)
                self.out.write(
                    focus.format_meter(**{**focus.format_dict, "ncols": ncols})
                )

                self.out.flush()

    def format_summaries(self) -> str:
        s = ""
        for i, instance in enumerate(self.instances):
            if i > 0:
                s += "|"
            if i == self.cur:
                s += "["
            s += self.format_summary(instance)
            if i == self.cur:
                s += "]"
        s += "| "
        return s

    def format_summary(self, instance):
        if instance.total is not None:
            return f"{instance.n/instance.total:.0%}"
        else:
            return f"{instance.n:d}"

    def update(self):
        with self.cv_update:
            self.cv_update.notify()

    def add(self, instance: "tqdm_managed"):
        with self.modify_lock:
            self.instances.append(instance)
            self.switch(len(self.instances) - 1)
        self.update()

    def remove(self, instance: "tqdm_managed"):
        with self.modify_lock:
            try:
                index = self.instances.index(instance)
                if self.cur == index:
                    self.switch(index - 1)
                del self.instances[index]
            except ValueError:
                print(f"WARNING: removing an instance that is not managed by me")
                pass
        self.update()

    def switch(self, to: Optional[int] = None):
        if len(self.instances) == 0:
            self.cur = -1

        now = time.time()
        if to is not None or now - self.last_switch > self.switch_interval:
            self.last_switch = now
            if to is None:
                to = self.cur + 1
            self.cur = to % len(self.instances)


stderr_pbar = PBarManager(out=sys.stderr)


class tqdm_managed(std_tqdm):
    def __init__(self, *args, **kwargs):
        self.manager = stderr_pbar
        super().__init__(*args, **kwargs)

        if self.disable:
            return

        self.manager.add(self)

    def close(self):
        if self.disable:
            return

        self.disable = True

        with self.get_lock():
            self.manager.remove(self)

    def clear(self, *_, **__):
        pass

    def display(self, *_, **__):
        self.manager.update()


tqdm = tqdm_managed



"""
This module assists the logging standard library.  The main functionality is:

* maintain two frequently used handlers: a stderr handler and a file handler, both with
  rich and customizable formats.
* `setup` method to attach them to the root logger.
* `get_logger` method to quickly create a logger with customized level.
"""
import logging
import sys
from logging import CRITICAL, DEBUG, ERROR, INFO, WARNING
from logging.handlers import RotatingFileHandler
from pathlib import Path
from typing import Optional, Union

__all__ = [
    "setup",
    "get_logger",
]


LOGGING_NAMESPACE = "su"

handler_stderr = logging.StreamHandler(sys.stderr)
handler_file = None


def setup(
    log_file: Optional[Union[str, Path]] = None,
    level_stderr: Union[int, str] = logging.INFO,
    level_file: Union[int, str] = logging.DEBUG,
    fmt_stderr: str = "[{asctime}{levelname[0]}]{name}: {message}",
    datefmt_stderr: str = "%H:%M:%S",
    fmt_file: str = "[{asctime}|{relativeCreated:.3f}|{levelname:7}]{name}: {message} [@{filename}:{lineno}|{funcName}|pid {process}|tid {thread}]",
    datefmt_file: str = "%Y-%m-%d %H:%M:%S",
    clear_handlers: bool = True,
    **kwargs_file: dict,
):
    """
    Setup the stderr and file handlers, and attach them to the root logger.

    :param log_file: the log file to use; if None, no file handler is created (and any existing one would be removed)
    :param level_stderr: the level filter of the stderr handler
    :param level_file: the level filter of the file handler
    :param fmt_stderr: the format of the stderr handler (with {} style)
    :param fmt_file: the format of the file handler (with {} style)
    :param clear_handlers: if True, remove all existing handlers of the root logger; otherwise, keep them as is
    :param kwargs_file: other optional kwargs to the file handler (RotatingFileHandler)
    """
    global handler_stderr, handler_file

    root_logger = logging.getLogger(LOGGING_NAMESPACE)
    root_logger.propagate = False
    # set to NOTSET+1, so that child logger's effective level can be correctly computed based on this logger
    root_logger.setLevel(logging.NOTSET + 1)

    if clear_handlers:
        root_logger.handlers = []

    # update the stderr handler
    handler_stderr.setLevel(level_stderr)
    handler_stderr.setFormatter(
        logging.Formatter(fmt_stderr, datefmt_stderr, style="{")
    )
    root_logger.addHandler(handler_stderr)

    # update the file handler
    root_logger.removeHandler(handler_file)

    if log_file is not None:
        kwargs_file.setdefault("maxBytes", 10_000_000)
        kwargs_file.setdefault("backupCount", 1)
        handler_file = RotatingFileHandler(log_file, **kwargs_file)
        handler_file.setLevel(level_file)
        handler_file.setFormatter(logging.Formatter(fmt_file, datefmt_file, style="{"))

        root_logger.addHandler(handler_file)
    else:
        handler_file = None


def get_logger(name: str, level: Union[int, str] = logging.NOTSET):
    """
    Get a logger with specified name and level.
    """
    logger = logging.getLogger(LOGGING_NAMESPACE + "." + name)
    logger.setLevel(level)
    return logger



from pathlib import Path
import re
from typing import *
from .. import LoggingUtils, IOUtils


class Macro:

    logger = LoggingUtils.get_logger("latex.Macro")

    T = TypeVar("T")
    def __init__(self,
                 key: str,
                 value: Optional[T] = None,
                 value_func: Optional[Callable[[Dict[str, "Macro"]], T]] = None,
                 tostring_func: Optional[Callable[[T], str]] = None):
        """
        Defines a latex macro, using "\DefMacro{key}{value}".
        :param key: the key of the macro.
        :param value: the value of the macro, can be None.
        :param value_func: the (lazily evaluated) function to get the value of the macro, can be None. The function will receive one argument of a dictionary of {macro.key, macro} of macros defined in the same file, and should return the evaluated value.
        :param tostring_func: the function to format the value to string, can be None. If set to None, __str__() will be used.
        """
        self.key: str = key
        self.value: Optional[Macro.T] = value
        self.value_func: Optional[Callable[[Dict[str, Macro]], Macro.T]] = value_func
        self.tostring_func: Optional[Callable[[Macro.T], str]] = tostring_func
        return

    def eval_content(self, macros_indexed: Dict[str, "Macro"]) -> str:
        """
        Evaluates the latex macro, and formats to a string that defines the macro (i.e., \DefMacro{key}{value}).
        If both value and value_func are defined, will use the value_func to update the value.

        :param macros_indexed: the indexed dictionary of {macro.key, macro}.
        :requires: not (self.value is None and self.value_func is None)
        :return: the string representation of the macro.
        """
        if self.value is None and self.value_func is None:
            raise ValueError("Cannot evaluate a macro without any value definition")
        # end if

        if self.value_func is not None:
            self.value = self.value_func(macros_indexed)
        # end if
        self.logger.info(f"Macro {self.key} evaluates to {self.value}")

        tostring_func = self.tostring_func if self.tostring_func is not None else str
        return f"\\DefMacro{{{self.key}}}{{{tostring_func(self.value)}}}"

    # Deprecated
    @classmethod
    def define(cls, key: str, value_fmt: Union[str, Any], *values, **values_items) -> "Macro":
        if len(values) != 0 or len(values_items) != 0:
            return Macro(key, value=value_fmt.format(*values, **values_items))
        else:
            return Macro(key, value=value_fmt)
        # end if

    def use(self) -> str:
        latex_line = f"\\UseMacro{{{self.key}}}"
        return latex_line

    RE_DEF_MACRO = re.compile(r"\\DefMacro{(?P<key>[^}]+)}{(?P<value>[^}]*)}")

    @classmethod
    def load_from_file(cls, file: Path) -> Dict[str, "Macro"]:
        """
        Loads the macros from a latex file.
        Will convert the value of the macros to int or float, if possible.

        :param file: the latex file.
        :return: the indexed dictionary of {macro.key, macro}.
        """
        macros_indexed: Dict[str, Macro] = dict()

        lines: List[str] = IOUtils.load(file, "txt").split("\n")
        for line in lines:
            match = cls.RE_DEF_MACRO.fullmatch(line.strip())
            if match is not None:
                key = match.group("key")
                value = match.group("value")

                # Try to convert to int, then (if failing) float.
                try:
                    value = int(value)
                except:
                    try:
                        value = float(value)
                    except:
                        pass
                # end try, try

                macros_indexed[key] = Macro(key, value)
            # end if
        # end for

        return macros_indexed



import copy
from typing import Any, Callable, Generic, Iterable, Optional, Tuple, TypeVar

TElem = TypeVar("TElem")
TValue = TypeVar("TValue")


class Trie(Generic[TElem, TValue]):
    """
    A Trie tree stores the mapping from sequences to values. Each item in the sequence is stored as a node in the tree, and sequences with the same prefix share the same node, to speed up the lookup.

    The tree is implemented recursively using dict, where each node is a dict of current node to its children. A special key `empty_elem` (by default empty string) is used to indicate a node has value.

    The value can be any object; the default `True` value is ok if the value is not important and only containment checking is needed.

    Using the default parameters of `empty_elem`, `join_func`, and `value` results in a classical trie tree for words.

    :param empty_elem: the empty element of this trie; by default empty string.
    :param join_func: the function to join a sequence of elements to a thing, used when reporting the sequence during traversing the trie; by default `"".join`.
    """

    MISSING = object()  # used to distinguish between None and missing values

    def __init__(
        self,
        empty_elem: TElem = "",
        join_func: Optional[Callable[[Iterable[TElem]], Any]] = lambda x: "".join(x),
    ):
        if join_func is None:
            join_func = lambda x: x
        self.empty_elem = empty_elem
        self.data = {}
        self.join_func = join_func

    def to_tree(self):
        """
        TODO: Converts the trie to a `seutil.ds.tree.Tree` (to be added).
        """
        raise NotImplementedError()

    def get(
        self, key: Iterable[TElem], default: Optional[TValue] = None
    ) -> Optional[TValue]:
        """
        Gets the value of the given key, or the default value if the key does not exist.
        """
        try:
            return self[key]
        except KeyError:
            return default

    def __getitem__(self, key: Iterable[TElem]) -> TValue:
        """
        Gets the value of the given key.
        :raises KeyError: if the key does not exist in the trie.
        """
        cur = self.data
        for c in key:
            if c not in cur:
                raise KeyError(f"Key {key} does not exist in the trie")
            cur = cur[c]
        if self.empty_elem not in cur:
            raise KeyError(f"Key {key} does not exist in the trie")
        return cur[self.empty_elem]

    def set(
        self, key: Iterable[TElem], value: TValue = True, exist_ok: bool = True
    ) -> TValue:
        """
        Sets the value of the given key.
        :param key: the key to set.
        :param value: the value to set, by default `True` if the value is not important.
        :param exist_ok: if the key already exists, whether to overwrite the value.
        :raises KeyError: if the key already exists and `exist_ok` is False.
        :return: the old value of the key, or `Trie.MISSING` if there was no old value.
        """
        cur = self.data
        for c in key:
            if c not in cur:
                cur[c] = {}
            cur = cur[c]
        if not exist_ok and self.empty_elem in cur:
            raise KeyError(f"Key {key} already exists in the trie")
        else:
            old_value = cur.get(self.empty_elem, self.MISSING)
            cur[self.empty_elem] = value
            return old_value

    def __setitem__(self, key: Iterable[TElem], value: TValue):
        self.set(key, value, exist_ok=True)

    def compute(
        self, key: Iterable[TElem], update_func: Callable[[Optional[TValue]], TValue]
    ) -> TValue:
        """
        Computes the value of the given key (potentially add/remove a key-value pair).
        Faster than get+set by avoiding searching for the key twice.
        :param key: the key to update.
        :param update_func: the function to update the value, which takes as input the old value (or `Trie.MISSING` if there was no old value), and returns the new value (or `Trie.MISSING` to indicate removing the key from trie).
        :return: the old value of the key, or `Trie.MISSING` if there was no old value.
        """
        cur = self.data
        for c in key:
            cur.setdefault(c, {})
            cur = cur[c]
        old_value = cur.get(self.empty_elem, None)
        cur[self.empty_elem] = update_func(old_value)
        if cur[self.empty_elem] == self.MISSING:
            del cur[self.empty_elem]
            self.prune()
        return old_value

    def remove(self, key: Iterable[TElem]) -> TValue:
        """
        Removes the key from the trie.
        :param key: the key to remove.
        :return: the old value of the key.
        :raises KeyError: if the key does not exist in the trie.
        """
        cur = self.data
        for c in key:
            if c not in cur:
                raise KeyError(f"Key {key} does not exist in the trie")
            cur = cur[c]
        if self.empty_elem not in cur:
            raise KeyError(f"Key {key} does not exist in the trie")
        value = cur[self.empty_elem]
        del cur[self.empty_elem]
        self.prune()
        return value

    def __delitem__(self, key: Iterable[TElem]):
        self.remove(key)

    def has_key(self, key: Iterable[TElem]) -> bool:
        """
        Checks if the key exists in the trie.
        """
        cur = self.data
        for c in key:
            if c not in cur:
                return False
            cur = cur[c]
        return self.empty_elem in cur

    def __contains__(self, key: Iterable[TElem]) -> bool:
        return self.has_key(key)

    def has_prefix(self, prefix: Iterable[TElem]) -> bool:
        """
        Checks if the prefix exists in the trie.
        """
        cur = self.data
        for c in prefix:
            if c not in cur:
                return False
            cur = cur[c]
        return True

    def __copy__(self):
        """
        Returns a shallow copy of the trie.
        """
        new_trie = Trie(self.empty_elem, self.join_func)

        def _copy(cur):
            new_cur = {}
            for k, v in cur.items():
                if k == self.empty_elem:
                    new_cur[k] = v
                else:
                    new_cur[k] = _copy(v)
            return new_cur

        new_trie.data = _copy(self.data)
        return new_trie

    def __deepcopy__(self, memo={}):
        """
        Returns a deep copy of the trie.
        """
        new_trie = Trie(self.empty_elem, self.join_func)
        new_trie.data = copy.deepcopy(self.data, memo)
        return new_trie

    def prune(self):
        """
        Prunes prefix nodes that lead to no value, usually after removal of a key.
        """

        def _prune(cur) -> Tuple[bool, list]:
            should_prune = True
            to_prune = []
            for c in cur:
                if c == self.empty_elem:
                    should_prune = False
                else:
                    should_prune_c, to_prune_c = _prune(cur[c])
                    if should_prune_c:
                        to_prune.append((cur, c))
                    else:
                        should_prune = False
                        to_prune += to_prune_c
            if should_prune:
                return True, []
            else:
                return False, to_prune

        _, to_prune = _prune(self.data)
        for cur, c in to_prune:
            del cur[c]

    def get_subtrie(self, prefix: Iterable[TElem]) -> "Trie[TElem, TValue]":
        """
        Gets the subtrie starting at the given prefix, with respect to this trie.
        If the prefix does not exist in this subtrie, an empty trie is returned.
        :param prefix: the prefix that the subtrie should start at.
        :return: the subtrie, which shares storage with the original trie (use `__copy__` or `__deepcopy__` to get a copy).
        """
        cur = self.data
        for c in prefix:
            if c not in cur:
                return Trie(self.empty_elem, self.join_func)
            cur = cur[c]
        new_trie = Trie(self.empty_elem, self.join_func)
        new_trie.data = cur
        return new_trie

    def get_subtrie_prefix_of(self, key: Iterable[TElem]) -> "Trie[TElem, TValue]":
        """
        Gets the subtrie that is bounded by the given key: all keys in the returned subtrie are prefix of the given key.
        If no prefix of the key exists in this subtrie, an empty trie is returned.
        :param key: the key that the subtrie should be bounded to.
        :return: the subtrie, which does not share structure, but shares values with the original trie (use `__deepcopy__` to get a deep copy).
        """
        cur = self.data
        new_trie = Trie(self.empty_elem, self.join_func)
        new_cur = new_trie.data
        for c in key:
            if self.empty_elem in cur:
                new_cur[self.empty_elem] = cur[self.empty_elem]

            if c not in cur:
                break
            else:
                new_cur[c] = {}
                new_cur = new_cur[c]
                cur = cur[c]

        if self.empty_elem in cur:
            new_cur[self.empty_elem] = cur[self.empty_elem]

        new_trie.prune()
        return new_trie

    def get_subtrie_using_elems(self, elems: Iterable[TElem]) -> "Trie[TElem, TValue]":
        """
        Gets the subtrie whose keys only use the given elements.
        :param elems: the elements that the subtrie can use.
        :return: the subtrie, which does not share structure, but shares values with the original trie (use `__deepcopy__` to get a deep copy).
        """
        new_trie = Trie(self.empty_elem, self.join_func)
        queue = [(self.data, new_trie.data)]
        while len(queue) > 0:
            cur, new_cur = queue.pop(0)
            if self.empty_elem in cur:
                new_cur[self.empty_elem] = cur[self.empty_elem]

            for c in elems:
                if c in cur:
                    new_cur[c] = {}
                    queue.append((cur[c], new_cur[c]))
        new_trie.prune()
        return new_trie

    def subtries(self) -> Iterable[Tuple[TElem, "Trie[TElem, TValue]"]]:
        """
        Iterates over all level-1 subtries of this trie.
        The returned subtrie shares storage with the original trie (use `__copy__` or `__deepcopy__` to get a copy).
        """
        for k, v in self.data.items():
            if k != self.empty_elem:
                subtrie = Trie(self.empty_elem, self.join_func)
                subtrie.data = v
                yield k, subtrie

    def items(self) -> Iterable[Tuple[Iterable[TElem], TValue]]:
        queue = [([], self.data)]
        while len(queue) > 0:
            prefix, cur = queue.pop()
            for k, v in cur.items():
                if k == self.empty_elem:
                    yield (self.join_func(prefix), v)
                else:
                    queue.append((prefix + [k], v))

    def keys(self) -> Iterable[Iterable[TElem]]:
        queue = [([], self.data)]
        while len(queue) > 0:
            prefix, cur = queue.pop()
            for k, v in cur.items():
                if k == self.empty_elem:
                    yield self.join_func(prefix)
                else:
                    queue.append((prefix + [k], v))

    def __iter__(self):
        return self.keys()

    def values(self) -> Iterable[TValue]:
        queue = [([], self.data)]
        while len(queue) > 0:
            prefix, cur = queue.pop()
            for k, v in cur.items():
                if k == self.empty_elem:
                    yield v
                else:
                    queue.append((prefix + [k], v))

    def __len__(self) -> int:
        return sum(1 for _ in self.values())

    def __str__(self):
        return f"Trie with {len(self)} elements"

    def __repr__(self) -> str:
        return f"Trie(empty_value={self.empty_elem}, data={self.data})"

    def overlaps_with(
        self, other: "Trie", self_has_value: bool = False, other_has_value: bool = False
    ) -> Iterable[Tuple[Iterable[TElem], "Trie", "Trie"]]:
        """
        Finds all subtries that overlap with the other trie (including the root subtries).
        :param other: the other trie to compare with.
        :param self_has_value: whether to only output overlaps where the self subtrie has a value.
        :param other_has_value: whether to only output overlaps where the other subtrie has a value.
        """
        queue = [(self.data, other.data, [])]
        while len(queue) > 0:
            self_node, other_node, prefix = queue.pop(0)

            # check if we should output this overlap
            if ((not self_has_value) or (self.empty_elem in self_node)) and (
                (not other_has_value) or (other.empty_elem in other_node)
            ):
                self_subtrie = Trie(self.empty_elem, self.join_func)
                self_subtrie.data = self_node
                other_subtrie = Trie(other.empty_elem, other.join_func)
                other_subtrie.data = other_node
                yield self.join_func(prefix), self_subtrie, other_subtrie

            # find all overlapping children
            for k, v in self_node.items():
                if k == self.empty_elem:
                    continue
                if k in other_node:
                    queue.append((v, other_node[k], prefix + [k]))
